"""Functions to read configuration."""

import configparser
import logging
import os
import re
from ast import literal_eval
from collections import OrderedDict, defaultdict
from glob import glob
from pathlib import Path
from typing import Any, Callable, ClassVar, Dict, List, Optional, Tuple, Union

from packaging.version import Version, parse

from . import __version__ as version
from ._toml import load as toml_load

SECTION = "odoo_scripts"
OTHER_SOURCES_KEY = "other_sources"
EXCLUDED_IN_SOURCES_KEY = "excluded_in_sources"
ODOO_7 = "odoo7"
"""Value for odoo_type representing Odoo 7.0"""
ODOO_8 = "odoo8"
"""Value for odoo_type representing Odoo 8.0"""
ODOO_9 = "odoo9"
""".. warning: This version is not handled"""
ODOO_10 = "odoo10"
"""Value for odoo_type representing Odoo 10.0"""
ODOO_11 = "odoo11"
"""Value for odoo_type representing Odoo 11.0"""
ODOO_12 = "odoo12"
""".. warning: This version is not handled"""
ODOO_13 = "odoo13"
"""Value for odoo_type representing Odoo 13.0"""
ODOO_14 = "odoo14"
"""Value for odoo_type representing Odoo 14.0"""
ODOO_15 = "odoo15"
"""Value for odoo_type representing Odoo 15.0"""
ODOO_16 = "odoo16"
"""Value for odoo_type representing Odoo 16.0"""
ODOO_17 = "odoo17"
"""Value for odoo_type representing Odoo 17.0"""
ODOO_18 = "odoo18"
"""Value for odoo_type representing Odoo 18.0"""

_logger = logging.getLogger(__name__)


class InvalidVersion(Exception):
    """The odoo_scripts version is not what is expected"""


class Configuration:
    """
    At the moment, only read from `setup.cfg`.
    """

    def __init__(
        self,
        path: str = ".",
        blacklist: Optional[str] = None,
        read: Optional[str] = None,
        parent_configuration: Optional["Configuration"] = None,
        registry: Optional[str] = None,
    ):
        """
        :param registry: use the provided registry instead of the one in the
        configuration file
        """
        self._expanded_configuration: Dict[Any, Configuration] = {}
        self.path: str = path
        setup_path: str = "setup.cfg"
        if path:
            setup_path = os.path.join(path, setup_path)
        config_parser = configparser.ConfigParser()
        if not os.path.exists(setup_path):
            _logger.warning("Missing %s", setup_path)
        else:
            config_parser.read(setup_path)
        section: Union[configparser.SectionProxy, dict] = (
            config_parser[SECTION] if SECTION in config_parser else {}
        )
        self._global_setup: Dict = {}
        global_setup_path = os.path.join(
            os.environ.get(
                "XDG_CONFIG_HOME", os.path.join(os.environ["HOME"], ".config")
            ),
            "odoo_scripts",
            "config.toml",
        )
        if os.path.exists(global_setup_path):
            with open(global_setup_path, "rb") as f:
                self._global_setup = dict(toml_load(f))

        # read expanded configurations information
        expanded_info: defaultdict[Any, Dict] = defaultdict(dict)
        for key in section:
            if key.startswith("expand."):
                remaining_key = key.split(".")[1:]
                expanded_info[remaining_key[0]][remaining_key[1]] = section.get(key)

        for name, value in expanded_info.items():
            value["parent_configuration"] = self
            if "path" not in value:
                _logger.error("Expansion without path for %s", name)
                raise Exception("Expansion without path for %s" % name)
            # put in relative_path the relative path between this and the parent.
            # this allows blacklisting odoo-mint/theme when expanding mint hr.
            if parent_configuration:
                # difference between this path and the parents
                relative_path = os.path.join(
                    os.path.relpath(
                        path,
                        parent_configuration.path if parent_configuration.path else "",
                    ),
                    value["path"],
                )
            else:
                relative_path = value["path"]
            if blacklist and re.match(blacklist, relative_path):
                _logger.info(
                    "Ignoring directory '%s' due to blacklisting",
                    os.path.join(path, value["path"]) if path else value["path"],
                )
            else:
                if path:
                    value["path"] = os.path.join(path, value["path"])
                if blacklist:
                    if "blacklist" in value:
                        # combine the two regexp
                        value["blacklist"] = "{}|{}".format(
                            value["blacklist"], blacklist
                        )
                    else:
                        value["blacklist"] = blacklist
                # read expanded configurations
                self._expanded_configuration[name] = Configuration(**value)

        def set_from_glob_values(
            base: Optional[str], glob_values: List[str], only_directories: bool
        ) -> set:
            result_set = set()
            for glob_value in glob_values:
                if base:
                    glob_value = os.path.join(base, glob_value)
                for file in glob(glob_value):
                    if not only_directories or os.path.isdir(file):
                        result_set.add(file)
            return result_set

        def toread(key: str) -> bool:
            """Return true if value of key is to be read from the file"""
            if read:
                return key in read.split(",")
            return True

        # those keys are path to files (glob compatible)
        self.modules: List[str]
        self.dependencies: List[str]
        self.other_sources: List[str]
        for key in (
            "modules",
            "dependencies",
            OTHER_SOURCES_KEY,
            EXCLUDED_IN_SOURCES_KEY,
        ):
            set_values = set()
            if toread(key):
                # add any expanded values
                for config in self._expanded_configuration.values():
                    set_values.update(getattr(config, key))
                # get the values from this configuration
                values = section.get(key, "").split()
                # XXX eventually add option to not include path
                set_values.update(
                    set_from_glob_values(path, values, key != OTHER_SOURCES_KEY)
                )
            if key == EXCLUDED_IN_SOURCES_KEY:
                self.excluded_paths = {Path(excluded) for excluded in set_values}
            else:
                setattr(self, key, list(set_values))

        def read_expanded(
            key: str,
            default_value=None,
            value_formatter: Optional[Callable[[str], Any]] = None,
        ):
            """Read from local version, or from expanded if value is found.
            Raise an exception if more than one value is found."""
            value = default_value
            if toread(key):
                found = False
                for config in self._expanded_configuration.values():
                    if hasattr(config, key):
                        if found:
                            raise Exception(
                                "Key %s to read in several expanded "
                                "configurations" % key
                            )
                        value = getattr(config, key)
                        found = True
                if key in section:
                    local_value = section.get(key, default_value)
                    if value_formatter is not None:
                        local_value = value_formatter(local_value)
                    if found and value != default_value and local_value != value:
                        raise Exception(
                            "Key %s in expanded configuration and locally set"
                            " with different values" % key
                        )
                    value = local_value
            if toread(key):
                setattr(self, key, value)

        self.min_version: Version
        read_expanded("min_version", parse("0"), parse)
        if hasattr(self, "min_version") and parse(version) < self.min_version:
            _logger.fatal("Odoo scripts %s is less than %s", version, self.min_version)
            raise InvalidVersion(
                "Odoo scripts {} is less than {}".format(version, self.min_version)
            )

        self.python_packages: Dict[str, Dict[str, Union[str, List[Tuple[str, str]]]]]
        if toread("python_packages"):
            self.python_packages = OrderedDict()
            # add any expanded values
            for config in self._expanded_configuration.values():
                if hasattr(config, "python_packages"):
                    for python_package in config.python_packages:
                        self.python_packages[python_package] = config.python_packages[
                            python_package
                        ]
            for key, value in literal_eval(
                section.get("python_packages", "{}")
            ).items():
                value_with_default = {
                    "dirs": [],
                    # XXX deprecate setting compile to a string value
                    "compile": value.get("compile", True) in (True, "true", "True"),
                    "exclude_in_sources": value.get("exclude_in_sources", False),
                }
                # previously authorized keys.
                if "mount" in value or "target" in value:
                    simple_options = (
                        value.get("mount", "."),
                        value.get(
                            "target",
                            os.path.join("odoo", "addons", os.path.basename(key)),
                        ),
                    )
                    value_with_default["dirs"].append(simple_options)
                if "dirs" in value:
                    for element in value["dirs"]:
                        value_with_default["dirs"].append(
                            (
                                element.get("mount", "."),
                                element.get(
                                    "target",
                                    os.path.join(
                                        "odoo",
                                        "addons",
                                        os.path.basename(
                                            element["mount"]
                                            if "mount" in element
                                            else key
                                        ),
                                    ),
                                ),
                            )
                        )
                elif "mount" not in value and "target" not in value:
                    value_with_default["dirs"].append(
                        (".", os.path.join("odoo", "addons", os.path.basename(key)))
                    )
                self.python_packages[key] = value_with_default

        self.addons_path = {os.path.dirname(module) for module in self.modules}
        """Set of directory containing modules"""

        def key_format(a_key: str) -> str:
            """Replace invalid characters in an attribute"""
            return a_key.replace("-", "_").replace(".", "_")

        # those keys are list of values
        self.module_list: List[str]
        self.module_list_tests: List[str]
        self.pg_extensions: List[str]
        for key in ("module_list", "module_list_tests", "pg.extensions"):
            value_list: List[str] = []
            formatted_key = key_format(key)
            if toread(key):
                # add any expanded values
                for config in self._expanded_configuration.values():
                    value_list.extend(getattr(config, formatted_key))
                # add local values
                value_list.extend(section.get(key, "").split())
            # then set it
            setattr(self, formatted_key, value_list)

        # those keys are single string values, no expand

        self.db_user: str
        self.db_password: str
        self.marabunta_db_user: str
        self.marabunta_db_password: str
        self.load_language: str
        self.duplicate_repo: str
        self.marabunta_migration_file: str
        for key in (
            "load-language",
            "duplicate_repo",
            "marabunta_migration_file",
        ):
            setattr(self, key_format(key), section.get(key, None))

        self.registry: Optional[str] = (
            section.get("registry", "registry.xcg.io") if registry is None else registry
        )
        project_path = os.path.realpath(".")
        self.image: str = section.get("image", os.path.basename(project_path))
        self.repository: str = f"{self.registry}/{self.image}"
        """Registry and image name combination, read from the configuration file"""
        self.local_tag: str = "dev"
        """Tag used for local operations"""
        # Does not use repository as it is not meant to be pushed
        self.local_image: str = self.image
        self.local_image_name: str = f"{self.local_image}:{self.local_tag}"
        """Full image name used for local operations (locally built, used for test and
        running)"""
        self.orus_api_token_required: bool
        read_expanded(
            "orus_api_token_required",
            False,
            lambda value: value in ["True", "true", True],
        )
        self.odoo_type: str
        read_expanded("odoo_type")
        # Fallback on ODOO_VERSION environment variable
        if not hasattr(self, "odoo_type") or self.odoo_type is None:
            if "ODOO_VERSION" in os.environ:
                self.odoo_type = (
                    f"odoo{os.environ['ODOO_VERSION'].split('.', maxsplit=1)[0]}"
                )
        if not hasattr(self, "odoo_type"):
            # if path is set, odoo_type might not be in the values to read
            # so no message
            if not path:
                _logger.warning("Missing odoo_type")
        elif self.odoo_type not in (
            ODOO_7,
            ODOO_8,
            ODOO_10,
            ODOO_11,
            ODOO_13,
            ODOO_14,
            ODOO_15,
            ODOO_16,
            ODOO_17,
            ODOO_18,
        ):
            if path:
                _logger.info(
                    "Unexpected odoo_type: %s in %s",
                    self.odoo_type,
                    self.path,
                )
            else:
                _logger.warning("Unexpected odoo_type: %s", self.odoo_type)
        for key in (
            "db_user",
            "db_password",
            "marabunta_db_user",
            "marabunta_db_password",
        ):
            option_value: Optional[str] = None
            if key in section:
                option_value = section[key]
            else:
                if key in self._global_setup:
                    option_value = self._global_setup[key]
                if hasattr(self, "odoo_type") and key in self._global_setup.get(
                    self.odoo_type, {}
                ):
                    option_value = self._global_setup[self.odoo_type][key]
            setattr(self, key_format(key), option_value)
        self.log_handler: List[str]
        self.pythondevmode: bool
        for key, default_value in (("log-handler", []), ("pythondevmode", True)):
            any_value: Any = default_value
            if key in self._global_setup:
                any_value = self._global_setup[key]
            if hasattr(self, "odoo_type") and key in self._global_setup.get(
                self.odoo_type, {}
            ):
                any_value = self._global_setup[self.odoo_type][key]
            setattr(self, key_format(key), any_value)

        self.postgresql_version: str
        read_expanded("postgresql_version", "9.6")
        self.start_py3o: bool = section.get("start_py3o", "no") in ("yes", "true")
        self.start_rednerd: bool = section.get("start_rednerd", "no") in (
            "yes",
            "true",
        )
        self.rednerd_version: str
        read_expanded("rednerd_version", "latest")
        self.rednerd_repository: Optional[str]
        read_expanded("rednerd_repository", None)
        self.__sources_modules: Optional[List[str]] = None

    @property
    def sources_modules(self) -> List[str]:
        """modules without those matching excluded_in_sources"""
        if self.__sources_modules is None:
            self.__sources_modules = list(
                filter(
                    lambda module: self.included(Path(module)),
                    self.modules,
                )
            )
        return self.__sources_modules

    def included(self, module_path: Path) -> bool:
        """
        :return: True if the module Path is to be included
        """
        return all(
            excluded_path not in module_path.parents
            for excluded_path in self.excluded_paths
        )


class Singleton(type):
    """Singleton metaclass. From
    https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """

    _instances: ClassVar[Dict] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Config(Configuration, metaclass=Singleton):
    """Singleton configuration."""

    def __init__(self):
        super().__init__()
