import logging
import os.path
from datetime import datetime
from typing import Collection, Optional, Set, Tuple

from git import Actor, Head, TagReference
from git import Repo as GitRepo
from git.exc import InvalidGitRepositoryError

from ._repo import Repo, TagExistsException, repo_kind

_logger = logging.getLogger(__name__)


class Git(Repo):
    """Represent a git repository"""

    def __init__(self, directory, git_repo: Optional[GitRepo] = None):
        super().__init__(directory)
        if git_repo is None:
            try:
                self.__git = GitRepo(directory)
            except InvalidGitRepositoryError:
                # try bare
                self.__git = GitRepo.init(directory, bare=True)
        else:
            self.__git = git_repo
        self.__orphan: bool = False
        self.__branch: Optional[str] = None
        self.__tags_to_push: Set[TagReference] = set()
        self.__tags_to_remove: Set[TagReference] = set()

    @property
    def kind(self) -> repo_kind:
        return "git"

    def remote_url(self, path: Optional[str] = None):
        remote = "origin" if path is None else path
        return self.__git.remote(remote)

    def branches(self):
        return self.__git.branches

    @property
    def ancestor_branches(self) -> Collection[str]:
        if self.__git.head.is_detached:
            return []
        # for each ancestor, look if it is a branch.
        # in git that only work if the ancestor branch has not changed since
        head_commits = {
            head.commit: head.name
            for head in self.__git.heads
            if head != self.__git.active_branch
        }
        ancestor_branches = []
        for commit in self.__git.iter_commits(self.__git.active_branch):
            if commit in head_commits:
                ancestor_branches.append(head_commits[commit])
        _logger.info("Ancestor branches: %s", ancestor_branches)
        return ancestor_branches

    @property
    def current_branch(self) -> str:
        if self.__git.head.is_detached:
            for head in self.__git.heads:
                # XXX seems overkill
                if self.__git.head.commit in self.__git.iter_commits(head.path):
                    return head.name
            # if here, we are in a commit not in the history of any branch
        return self.__git.active_branch.name

    @property
    def current_tags(self) -> Collection[str]:
        return [
            tag.name for tag in self.__git.tags if tag.commit == self.__git.head.commit
        ]

    @property
    def is_dirty(self) -> bool:
        return self.__git.is_dirty()

    @classmethod
    def clone_from(cls, url, to_path) -> "Repo":
        return Git(to_path, git_repo=GitRepo.clone_from(url, to_path))

    def switch(
        self, branch: str, ancestor_branches: Collection[str], create: bool = True
    ) -> bool:
        reference_name = f"origin/{branch}"
        _logger.debug("searching for reference %s", reference_name)
        branch_exists = False
        # search the branch
        for reference in self.__git.references:
            _logger.debug("found reference %s", reference.name)
            if reference.name == reference_name:
                self.__git.create_head(branch, reference_name)
                self.__git.heads[branch].checkout()

                branch_exists = True
                _logger.info(
                    "Found branch %s in %s (%s)",
                    reference.name,
                    self._directory,
                    reference,
                )
                break
        # try ancestors
        if not branch_exists:
            found = False
            for ancestor_branch in ancestor_branches:
                _logger.debug("Trying ancestor branch %s", ancestor_branch)
                found = self.switch(ancestor_branch, [], False)
                if found:
                    if create:
                        self.__git.create_head(branch, f"origin/{ancestor_branch}")
                        self.__git.heads[branch].checkout()
                        branch_exists = True
                    break
            # indicate to create the new branch (used only when commiting)
            self.__orphan = create and not found
            if self.__orphan:
                self.__git.head.set_reference(Head(self.__git, f"refs/heads/{branch}"))
                branch_exists = True
        # if create and not branch_exists:
        # will be used in Git::push
        self.__branch = branch
        return branch_exists

    def write_archival_txt(self, target):
        with open(
            os.path.join(target, ".git_archival.txt"), "w", encoding="utf-8"
        ) as f:
            f.write(
                self.__git.git.log(
                    "-1",
                    format="node: %H\n"
                    "node-date: %cI\n"
                    "describe-name: %(describe:tags)\n"
                    "ref-names: %D\n",
                )
            )

    def pre_switch_rm_files(self, target: str):
        # Need to clean up submodules **before** switching branches
        if self.__git.submodules:
            _logger.info("Cleaning up %d submodules", len(self.__git.submodules))
        for submodule in self.__git.submodules:
            submodule.remove()

    @classmethod
    def keep_on_rm(cls, filename: str) -> bool:
        return filename in (".git", ".gitmodules")

    def addremove(self):
        # also handle remove/renames
        self.__git.git.add(all=True)

    @property
    def has_changes(self) -> bool:
        return self.__git.is_dirty()

    def commit(
        self,
        message: str,
        author_date: Optional[datetime] = None,
        author: Optional[Tuple[str, str]] = None,
    ):
        self.__git.index.commit(
            message,
            author=None if author is None else Actor(author[0], author[1]),
            author_date=author_date,
        )

    def tag(self, tags: Collection[str], force: bool = False):
        existing_tags = {tag.name: tag for tag in self.__git.tags}
        for tag in tags:
            # check if tag is already set
            if tag in existing_tags:
                if existing_tags[tag].commit != self.__git.head.commit:
                    if force:
                        self.__tags_to_remove.add(existing_tags[tag])
                        self.__git.delete_tag(existing_tags[tag])
                    else:
                        raise TagExistsException(
                            "Tag already exists on another changeset"
                        )
                else:
                    # tag exist and is the same commit, do no tag and go on
                    continue
            self.__tags_to_push.add(self.__git.create_tag(tag))

    def push(self):
        # push commits
        _logger.info("Pushing %s", self.__branch)
        remote = self.__git.remote()
        # TODO how to detect errors?
        remote.push(refspec=f"{self.__branch}:{self.__branch}")
        # delete tags
        while self.__tags_to_remove:
            tag = self.__tags_to_remove.pop()
            _logger.info("deleting %s", tag.name)
            # XXX find if possible using other command
            self.__git.git.push("origin", tag.path, delete=True)
        # push tags
        tag_refspec = [tag.path for tag in self.__tags_to_push]
        self.__tags_to_push = set()
        if tag_refspec:
            _logger.info("Pushing tags %s", tag_refspec)
            remote.push(tag_refspec)
        else:
            _logger.info("No tags to push")

    @property
    def current_author_date(self) -> datetime:
        return self.__git.head.commit.authored_datetime

    @property
    def current_author(self) -> Tuple[str, str]:
        author = self.__git.head.commit.author
        if author is None:
            return "", ""
        return author.name or "", author.email or ""
