import logging
import os.path
from datetime import datetime
from sys import version_info
from typing import Collection, Optional, Tuple

import hglib

from ._repo import Repo, repo_kind

_logger = logging.getLogger(__name__)

if version_info < (3, 11, 0):
    from backports.datetime_fromisoformat import MonkeyPatch

    MonkeyPatch.patch_fromisoformat()


class Mercurial(Repo):
    """Represent a git repository"""

    def __init__(self, directory, hg_client: Optional[hglib.client.hgclient] = None):
        super().__init__(directory)
        if hg_client is None:
            self.__hg = hglib.open(directory)
        else:
            self.__hg = hg_client
        self.__new_branch = False

    @property
    def kind(self) -> repo_kind:
        return "hg"

    def remote_url(self, path: Optional[str] = None):
        path_byte = b"default" if path is None else path.encode()
        return self.__hg.paths(path_byte).decode()

    def branches(self):
        return self.__hg.branches()

    @property
    def current_branch(self) -> str:
        return self.__hg.identify(branch=True).strip().decode()

    @property
    def ancestor_branches(self) -> Collection[str]:
        # hg log -r "parents(min(branch(.)))" -T "{branch}\n"
        return (
            self.__hg.rawcommand(
                hglib.client.cmdbuilder(
                    b"log", r=b"parents(min(branch(.)))", template=b"{branch}"
                )
            )
            .decode()
            .split("\n")
        )

    @property
    def current_tags(self) -> Collection[str]:
        tags = (
            self.__hg.rawcommand(
                hglib.client.cmdbuilder(b"identify", template=b"{tags}")
            )
            .decode("UTF-8")
            .split(":")
        )
        # remove pseudo-tag
        if "tip" in tags:
            tags.remove("tip")
        return tags

    @property
    def is_dirty(self) -> bool:
        return bool(
            self.__hg.rawcommand(
                hglib.client.cmdbuilder(b"identify", template=b"{dirty}")
            ).decode("UTF-8")
        )

    @classmethod
    def clone_from(cls, url, to_path) -> "Repo":
        encoded_to_path = to_path.encode()
        hg_client = hglib.clone(source=url.encode(), dest=encoded_to_path)
        hg_client.open()
        return Mercurial(to_path, hg_client)

    def switch(
        self, branch: str, ancestor_branches: Collection[str], create: bool = True
    ) -> bool:
        branch_exists_in_duplicate = False
        for branch_name, num, id_ in self.__hg.branches():
            if branch_name.decode() == branch:
                self.__hg.update(rev=num)
                branch_exists_in_duplicate = True
                _logger.debug(
                    "Found branch %s in %s (%d, %s)",
                    branch_name.decode(),
                    self._directory,
                    num,
                    id_.decode(),
                )
                break
        if not branch_exists_in_duplicate:
            found = False
            for ancestor_branch in ancestor_branches:
                _logger.debug("Trying ancestor branch %s", ancestor_branch)
                found = self.switch(ancestor_branch, [], False)
                if found:
                    break
            if create and not found:
                self.__hg.update(b"tip")
                _logger.debug(
                    "Branch and ancestors branches not found in %s, updating to tip",
                    self._directory,
                )
        if create and not branch_exists_in_duplicate:
            # create the branch
            self.__hg.branch(name=branch.encode())
            branch_exists_in_duplicate = True
            self.__new_branch = True
        return branch_exists_in_duplicate

    def write_archival_txt(self, target):
        content = self.__hg.rawcommand(
            hglib.client.cmdbuilder(
                b"identify", template=b"repo: {id}\n", r=b"first(ancestors(.))"
            )
        )
        content += self.__hg.rawcommand(
            hglib.client.cmdbuilder(
                b"identify",
                template=b"""node: {id}
branch: {branch}\n""",
            )
        )
        tags = ":".join(self.current_tags)
        if tags:
            content += b"tag: %s" % tags.encode()
        else:
            content += self.__hg.rawcommand(
                hglib.client.cmdbuilder(
                    b"identify",
                    template=b"""latesttag: {latesttag}
latesttagdistance: {latesttagdistance}
changessincelatesttag: {changessincelatesttag}""",
                )
                # replace as done in docker_build_copy::_copy_packages
            ).replace(b"latesttag: null", b"latesttag: 0")
        with open(os.path.join(target, ".hg_archival.txt"), "wb") as f:
            f.write(content)

    @classmethod
    def keep_on_rm(cls, filename: str) -> bool:
        return filename in (".hg", ".hgtags")

    def addremove(self):
        # Add and remove file, use similarity to find renames
        self.__hg.addremove(similarity=80)

    @property
    def has_changes(self) -> bool:
        return bool(self.__hg.status())

    def commit(
        self,
        message: str,
        author_date: Optional[datetime] = None,
        author: Optional[Tuple[str, str]] = None,
    ):
        # author_date_bytes = None if author_date is None else author_date.encode()
        author__bytes = (
            None if author is None else f"{author[0]} <{author[1]}>".encode()
        )
        self.__hg.commit(message.encode(), date=author_date, user=author__bytes)

    def tag(self, tags: Collection[str], force: bool = False):
        self.__hg.tag(list(map(lambda s: s.encode(), tags)), force=force)

    def push(self):
        _logger.info("Pushing")
        self.__hg.push(newbranch=self.__new_branch)
        self.__new_branch = False

    @property
    def current_author_date(self) -> datetime:
        return datetime.fromisoformat(
            self.__hg.rawcommand(
                hglib.client.cmdbuilder(b"log", r=".", template=b"{date|isodate}")
            ).decode()
        )

    @property
    def current_author(self) -> Tuple[str, str]:
        return (
            self.__hg.rawcommand(
                hglib.client.cmdbuilder(
                    b"log", r=".", template=b"{author|person}\n{author|email}"
                )
            )
            .decode()
            .split("\n")
        )
