import logging
import os
import shutil
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Collection, Literal, Optional, Tuple

_logger = logging.getLogger(__name__)

repo_kind = Literal["git", "hg"]


class TagExistsException(Exception):
    """Exception indicating a tag already exists."""


class Repo(ABC):
    """Version System Repository abstraction"""

    def __init__(self, directory, **kwargs):  # pylint: disable=unused-argument
        super().__init__()
        self._directory = directory

    @property
    @abstractmethod
    def kind(self) -> repo_kind:
        """Return the CVS kind"""
        # inverse of what .__init__/get_repo uses in get_repo/clone_from

    @abstractmethod
    def remote_url(self, path: Optional[str] = None):
        """Return url of remote matching path/remote name"""

    @abstractmethod
    def branches(self):
        pass

    @property
    @abstractmethod
    def ancestor_branches(self) -> Collection[str]:
        """Return the ancestor branches name of the current branch"""

    @property
    @abstractmethod
    def current_branch(self) -> str:
        """Current branch of the repository"""

    @property
    @abstractmethod
    def current_tags(self) -> Collection[str]:
        """Current tags of the repository"""

    # TODO check if is_dirty can be replaced by has_changes
    @property
    @abstractmethod
    def is_dirty(self) -> bool:
        pass

    @classmethod
    @abstractmethod
    def clone_from(cls, url, to_path) -> "Repo":
        """Clone given repository and return a Repo matching it."""

    @abstractmethod
    def switch(
        self, branch: str, ancestor_branches: Collection[str], create: bool = True
    ) -> bool:
        """Update/switch to given branch."""
        return False

    @abstractmethod
    def write_archival_txt(self, target):
        """Write a .XXX_archival.txt file in target directory matching this repo
        values."""

    @abstractmethod
    def addremove(self):
        """Add and remove all files in the repository"""

    def pre_switch_rm_files(self, target: str):
        """Rm files before calling switch"""
        return

    def rm_files(self, target: str):
        """Delete all files in the repository"""
        _logger.info("Deleting target content")
        for filename in os.listdir(target):
            file_path = os.path.join(target, filename)
            if not self.keep_on_rm(filename):
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)

    @classmethod
    @abstractmethod
    def keep_on_rm(cls, filename: str) -> bool:
        """Return tree for file to keep when running rm_files"""

    @property
    @abstractmethod
    def has_changes(self) -> bool:
        return False

    @abstractmethod
    def commit(
        self,
        message: str,
        author_date: Optional[datetime] = None,
        author: Optional[Tuple[str, str]] = None,
    ):
        """Commit changes"""

    @abstractmethod
    def tag(self, tags: Collection[str], force: bool = False):
        """Add tag on current revision"""

    @abstractmethod
    def push(self):
        """Push the repository changes (commits and tags)"""

    @property
    @abstractmethod
    def current_author_date(self) -> datetime:
        """Author date of the commit in current revision"""

    @property
    @abstractmethod
    def current_author(self) -> Tuple[str, str]:
        """Author of the commit in current revision"""
