#!/usr/bin/env python3
"""Tools to help build an image"""

import argparse
import functools
import logging
import os
import sys
import tempfile
from collections.abc import Mapping
from multiprocessing import Pool
from subprocess import DEVNULL, CalledProcessError, check_call, check_output, run
from typing import List, Optional, Tuple, Union

import hglib

from odoo_scripts._logpipe import LogPipe

from ._config import Config
from ._cvs import get_repo
from ._parsing import apply, basic_parser
from ._toml import load
from .list_modules import list_modules
from .storage import get_orus_api_token

_logger = logging.getLogger(__name__)

__version__ = "2.6.2"
__date__ = "2020-06-30"
__updated__ = "2024-10-10"

_ODOO_MODULES_DIR = "odoo_modules"
_PYTHON_PACKAGES_DIR = "python_packages"


def add_build_copy_options(parser: argparse.ArgumentParser):
    # TODO ajouter une option pour ne pas ignorer la version locale d’un module
    #   et une autre pour ne rien télécharger (--no-download).
    #   Éventuellement une autre pour désactiver tout cache.
    parser.add_argument(
        "--orus-api-token",
        help="Token to access packages on orus.io",
        default=None,
    )
    parser.add_argument(
        "--empty",
        help="Do not copy any module or package",
        action="store_true",
        dest="empty",
        default=False,
    )


# XXX Using something else than redone arguments would be better
def get_build_copy_options(namespace: argparse.Namespace) -> List[str]:
    """Return build options usable in :func:`build_local_image`."""
    result = []
    if namespace.orus_api_token:
        result.append("--orus-api-token")
        result.append(namespace.orus_api_token)
    return result


class Configuration(Mapping):
    def __init__(self, namespace):
        super().__init__()
        self.orus_api_token = (
            namespace.orus_api_token
            if namespace.orus_api_token
            else get_orus_api_token()
        )
        self.empty = namespace.empty

    def __len__(self):
        return 1

    def __getitem__(self, item):
        return getattr(self, item)

    def __iter__(self):
        yield from ("orus_api_token", "empty")


def __parser():
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = f"""{program_shortdesc}

      Created by Vincent Hatakeyama on {__date__}.
      Copyright 2020, 2021, 2023 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """
    parser = basic_parser(program_license, program_version_message)
    add_build_copy_options(parser)
    return parser


def main(argv: Optional[List[str]] = None) -> int:
    """Copy modules for a build, callable version that parses arguments"""
    parser = __parser()
    nmspc = parser.parse_args(argv)
    apply(nmspc)
    copy(**Configuration(nmspc))
    return 0


def _target_path(path: str) -> str:
    base = os.path.basename(os.path.realpath("."))
    if path == ".":
        return base
    return os.path.join(
        os.path.dirname(os.path.join(base, path)).replace(os.path.sep, "-"),
        os.path.basename(path),
    )


def copy(
    build_package_locally: Optional[bool] = None,
    orus_api_token: Optional[str] = None,
    download: bool = True,
    empty: bool = False,
):
    """Copy modules for a build

    :param build_package_locally: build package that have package compile set to True
      rather than just copy them. (in CI, build package to reduce artifact size, locally
      default to copying them)
    :param orus_api_token: token with read_api rights on orus.io
    :param download: indicate if downloading packages should be attempted
    :param empty: if true, do not copy any files
    """
    c = Config()
    for target in (_ODOO_MODULES_DIR, _PYTHON_PACKAGES_DIR):
        if os.path.exists(target):
            os.rmdir(target)
        os.mkdir(target)
    list_modules()
    if empty:
        # Add empty requirements
        with open(
            os.path.join(_PYTHON_PACKAGES_DIR, "requirements"), "wt", encoding="utf-8"
        ) as f:
            f.write("")
        return

    modules = [os.path.realpath(module) for module in c.modules]
    target = _ODOO_MODULES_DIR
    cmd = [
        "rsync",
        "--delete",
        "--cvs-exclude",
        "--include=core",
        "--exclude=.hg*",
        "--exclude=.git*",
        "--links",
        "--exclude='*.pyc'",
        "-r",
        "--times",
        *modules,
        target,
    ]
    _logger.debug(" ".join(cmd))
    check_call(cmd, stdout=DEVNULL)

    target = _PYTHON_PACKAGES_DIR
    packages_to_compile = []
    packages_to_copy = []
    requirements: List[str] = []
    ci_mode = "CI" in os.environ and os.environ["CI"].lower() == "true"
    # Try to include the CVS information in target so that setuptools-scm and
    # setuptools-odoo work as expected
    for package, options in c.python_packages.items():
        if not os.path.exists(package):
            raise Exception(
                "Missing %s referenced in configuration. Forgot to vendor?" % package
            )
        if (options["compile"] and ci_mode) or build_package_locally:
            packages_to_compile.append(package)
        else:
            packages_to_copy.append(package)
    if download:
        downloaded_packages = []
        _do = functools.partial(
            _search_and_download_package, orus_api_token=orus_api_token, target=target
        )
        for package_group in (packages_to_compile, packages_to_copy):
            if package_group:
                with Pool() as p:
                    results = p.map(_do, package_group)
                for result in results:
                    if result is not None:
                        package, required = result
                        package_group.remove(package)
                        downloaded_packages.append(package)
                        requirements.extend(required)
        if downloaded_packages:
            _logger.info("%d packages downloaded", len(downloaded_packages))
    if packages_to_compile:
        _logger.info("Compiling %d packages", len(packages_to_compile))
        with Pool() as p:
            compiled_requirements = p.map(
                functools.partial(_compile_package, target=target), packages_to_compile
            )
        for reqs in compiled_requirements:
            requirements.extend(reqs)
    if packages_to_copy:
        _logger.info("Copying %d packages", len(packages_to_copy))
        # Non multiprocess version
        # copied_requirements = []
        # for package in packages_to_copy:
        #     copied_requirements.append(_copy_packages(package, target))
        # Multiprocess version
        with Pool() as p:
            copied_requirements = p.map(
                functools.partial(_copy_packages, target=target), packages_to_copy
            )
        for copy_result in copied_requirements:
            requirements.extend(copy_result)

    # Write requirements file
    # pip needs to be run with pip install -r <target>/requirements for the paths in the
    # file to be valid
    with open(os.path.join(target, "requirements"), "wt", encoding="utf-8") as f:
        f.write("\n".join(requirements))

    # copy setup files to odoo_setup
    source = "setup/"
    target = "odoo_setup"
    if os.path.exists(source):
        _logger.info("Copying Odoo %s files to %s", source, target)
        if not os.path.exists(target):
            os.mkdir(target)
        cmd = [
            "rsync",
            "--delete",
            "--cvs-exclude",
            "--include=core",
            "--exclude=.hg*",
            "--exclude=.git*",
            "--copy-links",
            "--exclude='*.pyc'",
            "-r",
            "--times",
            *source,
            *target,
        ]
        _logger.debug(" ".join(cmd))
        check_call(cmd)


# TODO changer les arguments pour utiliser un nom différent que package
def _find_cvs_parent(package: str) -> Optional[Tuple[str, str]]:
    """Find the parent containing a CVS root.
    If no CVS root is found (beside the one of the super project), return the package
    itself.
    """
    cvs_parent = package
    package_path_in_target = ""
    while not (
        os.path.exists(os.path.join(cvs_parent, ".git"))
        or os.path.exists(os.path.join(cvs_parent, ".hg"))
        or os.path.exists(os.path.join(cvs_parent, ".hg_archival.txt"))
        or os.path.exists(os.path.join(cvs_parent, ".git_archival.txt"))
    ):
        if os.path.dirname(cvs_parent) in ("", "."):
            # stop at one level in the super project, if no CVS info found, only copy
            # target
            return None
        package_path_in_target = os.path.join(
            os.path.basename(cvs_parent), package_path_in_target
        )
        cvs_parent = os.path.dirname(cvs_parent)
    return cvs_parent, package_path_in_target


# TODO revoir l’argument avec la fonction, ce n’est pas l’idéal. peut-être pas le choix
def _copy_packages(
    package,
    target: str,
    compute_path=lambda _cvs_parent, package: _target_path(package),
    can_use_archive: Optional[bool] = None,
) -> List[str]:
    # can_use_archive has a backward compatible value: None
    # Before the argument was added, it was False for git, and true for mercurial.
    archive_success: bool = False
    result = _find_cvs_parent(package)
    if result is not None:
        cvs_parent, package_path_in_target = result
        this_package_target = os.path.join(target, compute_path(cvs_parent, package))
        # creates parent directory of target
        os.makedirs(os.path.dirname(this_package_target), exist_ok=True)
        # Use hg archive
        if os.path.exists(os.path.join(cvs_parent, ".hg")) and can_use_archive in (
            True,
            None,
        ):
            _logger.info(
                "Package in %s: mercurial archive to %s",
                cvs_parent,
                this_package_target,
            )
            repo = get_repo(cvs_parent)
            client = hglib.open(cvs_parent)
            try:
                # TODO use repo instead of client
                client.archive(this_package_target.encode())
                archive_success = True
            except hglib.error.CommandError:
                _logger.warning(
                    "Archive creation failed for %s", cvs_parent, exc_info=True
                )
            if archive_success:
                # fix for setuptools-scm/hatchling that do not like when the file has
                # latesttag: null
                check_call(
                    [
                        "sed",
                        "-i",
                        "-s",
                        "s,latesttag: null,latesttag: 0,",
                        os.path.join(this_package_target, ".hg_archival.txt"),
                    ]
                )
                if repo.is_dirty:
                    _logger.info("Local changes in %s, applying them", cvs_parent)
                    logger = _logger.getChild("patch")
                    with tempfile.TemporaryFile() as fp:
                        # TODO use repo instead of client
                        fp.write(client.diff())
                        fp.seek(0)
                        check_call(
                            ["patch", "-p1", "-f"],
                            cwd=this_package_target,
                            stdin=fp,
                            stdout=LogPipe(logger, logging.INFO),  # type: ignore[arg-type]
                            stderr=LogPipe(logger, logging.ERROR),  # type: ignore[arg-type]
                        )
        if os.path.exists(os.path.join(cvs_parent, ".git")):
            if can_use_archive:
                _logger.info(
                    "Package in %s: Copying %s to %s (no cvs)",
                    package,
                    cvs_parent + os.path.sep,
                    this_package_target,
                )
                _copy_package(
                    cvs_parent + os.path.sep,
                    this_package_target,
                    cvs_parent == ".",
                    True,
                )
                get_repo(cvs_parent).write_archival_txt(this_package_target)
                archive_success = True
    else:
        cvs_parent, package_path_in_target = package, ""
        # XXX refactorize with code above
        this_package_target = os.path.join(target, compute_path(cvs_parent, package))
        # creates parent directory of target
        os.makedirs(os.path.dirname(this_package_target), exist_ok=True)

    # if archive failed (either not a hg, or command failed due to empty content)
    if not archive_success:
        _logger.info(
            "Package in %s: Copying %s to %s",
            package,
            cvs_parent + os.path.sep,
            this_package_target,
        )
        _copy_package(
            cvs_parent + os.path.sep, this_package_target, cvs_parent == ".", False
        )
    return [os.path.join(".", this_package_target, package_path_in_target)]


def _compile_package(package_to_compile, target) -> List[str]:
    versions = _find_packages_version(package_to_compile, tags_only=False)
    content = []
    requirements = []
    if versions:
        cachedirname = compiled_cache_dir(versions[0][0], versions[0][1])
        if not os.path.exists(cachedirname):
            os.makedirs(cachedirname)
        content = os.listdir(cachedirname)
    if not content:
        package_path = os.path.join(".", package_to_compile)
        with tempfile.TemporaryDirectory() as tmpdirname:
            try:
                # TODO pip wheel semble plus rapide
                # build is not usable directly so use a subprocess.
                # Disable flake8, this is just to test that the package is installed.
                import build  # type: ignore[import] # noqa: F401

                cmd = [
                    sys.executable,
                    "-m",
                    "build",
                    "--outdir",
                    tmpdirname,
                    package_path,
                ]
                _logger.debug(" ".join(cmd))
                check_call(cmd)
            except ImportError:
                # pip is not usable directly, it is indicated in its doc to use
                # subprocess instead.
                cmd = [
                    sys.executable,
                    "-m",
                    "pip",
                    "wheel",
                    "--no-deps",
                    "-w",
                    tmpdirname,
                    "--ignore-requires-python",
                    package_path,
                ]
            _logger.debug(" ".join(cmd))
            check_call(cmd)
            content = os.listdir(tmpdirname)
            for file in content:
                # LINUX
                check_call(["cp", os.path.join(tmpdirname, file), target])
                if versions:
                    # LINUX
                    check_call(["cp", os.path.join(target, file), cachedirname])
    else:
        for file in content:
            # LINUX
            check_call(["cp", os.path.join(cachedirname, file), target])
    for file in content:
        requirements.append(os.path.join(".", target, file))

    try:
        # Disable flake8, this is just to test that the package is installed
        import twine  # type: ignore[import] # noqa: F401

        cmd = [sys.executable, "-m", "twine", "check", f"{target}/*"]
        check_call(cmd)
    except ImportError:
        _logger.warning("No twine, no check done")
    return requirements


def _search_and_download_package(package, orus_api_token, target) -> Union[None, Tuple]:
    potential_packages: List[Tuple[str, str]] = _find_packages_version(package)
    if potential_packages:
        return _download_package(package, orus_api_token, potential_packages, target)
    return None


def _download_package(
    package,
    orus_api_token,
    package_to_download: List[Tuple[str, str]],
    target,
):
    # Download
    pip_download_command = [
        "pip",
        "download",
        "--no-deps",
        "--ignore-requires-python",
        "--only-binary",
        ":all:",
        xcg_index_url(orus_api_token),
    ]
    requirements = []
    for package_name, version in package_to_download:
        try:
            download_target = cache_dir(package_name, version)
            content = []
            if os.path.isdir(download_target):
                content = os.listdir(download_target)
            if not content:
                _logger.debug("Downloading %s with pip", package_name)
                check_call(
                    [
                        *pip_download_command,
                        f"{package_name} =={version}",
                        "--dest",
                        download_target,
                    ]
                )
                content = os.listdir(download_target)
            else:
                _logger.debug("Reusing cache for %s", package_name)
            for file in content:
                # LINUX
                check_call(["cp", "-a", os.path.join(download_target, file), target])
                requirements.append(os.path.join(".", target, file))
            return package, requirements
        except CalledProcessError:
            # Ignore issues, mainly when the package version does not exist
            pass
    return None


def _find_packages_version(package, tags_only=True) -> List[Tuple[str, str]]:
    # Try to download any package with a clear name
    toml_file = os.path.join(package, "pyproject.toml")
    if os.path.exists(toml_file):
        _logger.debug("Reading %s", toml_file)
        # TODO semble long, mettre en cache avec la date du fichier pour éviter
        #  de relire à chaque fois?
        with open(toml_file, "rb") as f:
            package_name = load(f).get("project", {}).get("name", "")
        if package_name:
            _logger.debug("Package name %s for %s", package_name, package)
            # try to find out version
            hg_root = run(["hg", "root"], cwd=package, capture_output=True)
            if not hg_root.returncode:
                # removes trailing \n
                hg_root_dir = hg_root.stdout[:-1]
                _logger.debug(
                    "Directory %s is a part of a mercurial repository",
                    hg_root_dir.decode("UTF-8"),
                )
                client = hglib.open(hg_root_dir)
                # TODO check if is dirty
                if tags_only:
                    tags = client.identify(tags=True).split()
                    package_list = [
                        (package_name, tag.decode("UTF-8"))
                        for tag in tags
                        if tag != b"tip"
                    ]
                else:
                    version = client.identify(id=True).decode("UTF-8")[:-1]
                    package_list = [(package_name, version)]
                if package_list:
                    _logger.debug("Matches for %s: %s", package, package_list)
                    return package_list

            elif (
                # TODO use gitpython
                check_output(["git", "rev-parse", "--is-inside-work-tree"], cwd=package)
                == "true"
            ):
                repo = get_repo(package)
                _logger.debug("Directory is a part of a git repository")
                # TODO 1. check if is dirty
                if repo.is_dirty:
                    _logger.debug("Repository is clean")
                    #  2. if clean, check tag
                    if tags_only:
                        tag = check_output(
                            # TODO use gitpython
                            ["git", "describe", "--exact-match", "HEAD"],
                            cwd=package,
                        )
                        if tag:
                            # Ignore local python version, this will be installed in an
                            # image.
                            # TODO changer pour prendre la version du python du super
                            #  projet. Probablement pas renseigné où que ce soit.
                            return [(package_name, tag.decode("UTF-8"))]
                        _logger.debug("No tag found")
                    else:
                        return [
                            (
                                package_name,
                                check_output(
                                    # TODO use gitpython
                                    ["git", "rev-parse", "HEAD"],
                                    cwd=package,
                                ).decode("UTF-8"),
                            )
                        ]
                else:
                    _logger.debug("Repository is dirty")
            else:
                _logger.debug("Directory is not part of any kind of known repository")
            # handle .hg_archival.txt and .git_archival.txt later
    return []


# TODO rename, this copies a directory, not just a package
def _copy_package(src, dest, exclude_super: bool, exclude_csv: bool):
    cmd = [
        "rsync",
        "--delete",
        "--include=core",
        "--links",
        "--exclude='*.pyc'",
        "--exclude=__pycache__",
        "--exclude='*.orig'",
        "-r",
        "--times",
        src,
        dest,
    ]
    if exclude_super:
        cmd.extend(
            [
                f"--exclude={_PYTHON_PACKAGES_DIR}/",
                f"--exclude={_ODOO_MODULES_DIR}/",
            ]
        )
    if exclude_csv:
        cmd.extend(
            [
                "--cvs-exclude",
            ]
        )
    _logger.debug(" ".join(cmd))
    check_call(cmd)


def xcg_index_url(token: Optional[str] = None):
    if token is not None:
        return (
            f"--index-url=https://__token__:{token}@orus.io/api/v4/groups/9/-/packages/"
            f"pypi/simple"
        )
    return "--extra-index-url=https://pypi.xcg.io/simple"


def base_cache_dir() -> str:
    """Return base cache directory"""
    return os.environ.get(
        "ODOO_SCRIPTS_CACHE_DIR",
        os.path.join(
            os.environ.get(
                "XDG_CACHE_HOME", os.path.join(os.environ["HOME"], ".cache")
            ),
            "odoo-scripts",
        ),
    )


def cache_dir(package: str, version: str, python_version: str = "unknown") -> str:
    """Return cache dir for package and version"""
    return os.path.join(base_cache_dir(), "packages", python_version, package, version)


def compiled_cache_dir(package: str, cvs_hash: str) -> str:
    """Return cache dir for package and version"""
    return os.path.join(base_cache_dir(), "compiled-packages", package, cvs_hash)


if __name__ == "__main__":
    sys.exit(main())
