#!/usr/bin/env python3
"""Build documentation in a docker"""

import argparse
import logging
import os
import sys
from typing import List, Optional

from ._config import Config
from ._docker_client import DockerClient, get_volumes
from ._parsing import apply, basic_parser
from .docker_build import add_build_options, build_local_image, get_build_options

__version__ = "1.0.1"
__date__ = "2022-11-10"
__updated__ = "2022-12-05"

_logger = logging.getLogger(__name__)

_DIR_DEST = "build_doc_dir"


def __parser():
    program_version = __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (
        program_version,
        program_build_date,
    )
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = """%s

      Created by Vincent Hatakeyama on %s.
      Copyright 2022 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """ % (
        program_shortdesc,
        str(__date__),
    )
    parser = basic_parser(program_license, program_version_message)
    return parser


def parser_add_builddoc(parser: argparse.ArgumentParser):
    """Add isort option (--isort/--no-isort) to the given parser.
    Defaults to not using isort."""
    parser.add_argument(
        help="Create documentation on given directory (relative path, for example "
        "github.com/OCA/server-auth)",
        dest=_DIR_DEST,
    )


def apply_build_doc(namespace: argparse.Namespace, config: Optional[Config] = None):
    """Run isort if the option was set."""
    if _DIR_DEST in namespace:
        return build_doc(
            directory=getattr(namespace, _DIR_DEST),
            config=config,
        )
    return 0


def build_doc(directory: str = "", config: Optional[Config] = None):
    """Run isort"""
    if config is None:
        config = Config()

    # determine image to use based on odoo version
    repository = config.local_image
    tag = config.local_tag

    pwd = os.environ["PWD"]
    volumes, path = get_volumes(pwd, directory, write=True)
    _logger.info("Building documentation in %s", path)
    return DockerClient.run(
        repository,
        tag,
        {
            "entrypoint": "make",
            "user": os.getuid(),
            "command": ["html"],
            "volumes": volumes,
            "workdir": path,
        },
        False,
    )


def main(argv: Optional[List[str]] = None) -> int:
    """Copy modules for a build, callable version that parses arguments"""
    parser = __parser()
    add_build_options(parser)
    parser_add_builddoc(parser)
    nmspc = parser.parse_args(argv)
    config = Config()
    apply(nmspc)
    build_local_image(get_build_options(nmspc))
    return apply_build_doc(nmspc, config=config)


if __name__ == "__main__":
    sys.exit(main())
