#!/usr/bin/env python3
"""Launch a docker image but bind the local directory inside the container
and define the module path automatically.
"""

import atexit
import logging
import os
import pwd
import socket
import sys
import tempfile
from argparse import ArgumentParser
from configparser import ConfigParser
from contextlib import closing
from enum import Enum
from subprocess import call, check_output
from typing import Any, Dict, List, Optional

import hglib
from psycopg2 import OperationalError, connect
from psycopg2.errors import DuplicateDatabase, DuplicateObject
from python_on_whales import docker

from ._config import (
    ODOO_7,
    ODOO_8,
    ODOO_9,
    ODOO_10,
    ODOO_11,
    ODOO_12,
    ODOO_13,
    ODOO_14,
    ODOO_15,
    Config,
)
from ._docker_client import (
    CaptureDockerExitCode,
    DockerClient,
    Mount,
    anthem_mounts,
    modules_mount,
    mount,
    odoo_source_mounts,
)
from ._parsing import apply, basic_parser
from ._toml import load
from .docker_build import add_build_options, build_local_image, get_build_options
from .docker_flake8 import apply_flake8, parser_add_flake8_group
from .docker_isort import apply_isort, parser_add_isort_group
from .docker_postgresql import POSTGRES_PASSWORD, docker_run_postgresql
from .docker_py3o import start_py3o
from .docker_pylint import apply_pylint, parser_add_pylint_group
from .docker_redis import DockerRedis
from .docker_rednerd import DockerRednerd, get_default_redner_database

# TODO auto create list of module

_logger = logging.getLogger(__name__)

__version__ = "3.18.1"
__date__ = "2017-08-11"
__updated__ = "2024-06-21"


class Command(Enum):
    # represents odoo --help
    HELP = "help"
    # represents running the odoo-bin server
    RUN = "run"
    # represents running the odoo-bin populate command
    POPULATE = "populate"
    # represent running an anthem song
    ANTHEM = "anthem"


def __parser(project_name: str) -> ArgumentParser:
    """Return a parser for docker_dev_start"""
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = f"""{program_shortdesc}

  Created by Vincent Hatakeyama on {__date__}.
  Copyright 2017-2024 XCG Consulting. All rights reserved.

  Licensed under the MIT License

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
"""

    # TODO add a way to store configuration options in a project file
    # Argument parsing
    parser = basic_parser(description=program_license, version=program_version_message)
    parser.add_argument(
        "--name",
        help="Container name to use [default: %(default)s]",
        default=project_name,
    )
    parser.add_argument(
        "--db_user", help="Database user [default: %(default)s]", default=None
    )
    parser.add_argument(
        "--db_password",
        help="Database user password [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--db_host",
        help="database host address",
    )
    command = parser.add_mutually_exclusive_group()
    start = command.add_argument_group()
    start.add_argument(
        "--max-cron-threads",
        help="max cron threads [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "-d",
        "--database",
        help="Database [default: %(default)s]",
        default=None,
    )
    parser.add_argument("-p", "--dbport", help="Change database port", default=None)
    group = start.add_mutually_exclusive_group()
    group.add_argument(
        "-u",
        "--update",
        help="Module to update (will also set --i18n-overwrite)"
        " [default: %(default)s]\n"
        "Options --update, --test and --test-default"
        " cannot be used at the same time",
        default=None,
    )
    group.add_argument(
        "--install-test",
        help="Modules to install for tests (will also set --log-level=test)"
        " [default: %(default)s]\n"
        "Options --update, --test, --install-test and --test-default"
        " cannot be used at the same time",
        default=None,
    )
    group.add_argument(
        "-t",
        "--test",
        help="Modules to test (will also set --log-level=test)"
        " [default: %(default)s]\n"
        "Options --update, --test and --test-default"
        " cannot be used at the same time",
        default=None,
    )
    group.add_argument(
        "--test-default",
        help="Test all modules in module_list_test"
        " [default: %(default)s]\n"
        "Options --update, --test and --test-default"
        " cannot be used at the same time",
        action="store_true",
    )
    start.add_argument(
        "--test-tags",
        help="Option sent to odoo [default: %(default)s]",
        default=None,
    )
    populate = command.add_argument_group()
    populate.add_argument(
        "--populate-model",
        help="Populate the indicated model",
        default=None,
    )
    populate.add_argument(
        "--populate-size",
        help="Populate size",
        default=None,
    )
    start.add_argument(
        "-i",
        "--install",
        help="Module to install [default: %(default)s]",
        default=None,
    )
    start.add_argument(
        "--without-demo",
        help="Module to avoid including demo data [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--create-user",
        help="Create user in database (using $user) [default: %(default)s]",
        action="store_true",
    )
    start.add_argument(
        "--install-default",
        help=(
            "Append the module_list from setup.cfg to the modules to install"
            " [default: %(default)s]"
        ),
        action="store_true",
    )
    parser.add_argument(
        "--start-postgresql",
        help="Start a Postgresql docker",
        action="store_true",
    )
    parser_add_flake8_group(parser)
    parser_add_isort_group(parser)
    parser_add_pylint_group(parser)
    add_build_options(parser)
    command.add_argument(
        "--odoo-help",
        help="Pass --help to odoo binary [default: %(default)s]",
        action="store_true",
    )
    start.add_argument(
        "--stop-after-init",
        help="Pass --stop-after-init to odoo binary [default: %(default)s]",
        action="store_true",
    )
    build_group = parser.add_mutually_exclusive_group()
    build_group.add_argument(
        "--no-build",
        help="Do not build locally [default: %(default)s]",
        action="store_true",
    )
    build_group.add_argument(
        "--force-build",
        help="Force new local build [default: %(default)s]",
        action="store_true",
    )
    network_group = parser.add_mutually_exclusive_group()
    network_group.add_argument(
        "--host-network",
        help="Use host network [default: %(default)s]",
        action="store_true",
        default=True,
    )
    network_group.add_argument(
        "--docker-network",
        help=(
            "Use docker network (works better with"
            " python-netifaces installed) [default: %(default)s]"
        ),
        action="store_false",
        dest="host_network",
        default=False,
    )
    dev_option_group = start.add_mutually_exclusive_group()
    dev_option_group.add_argument(
        "--no-dev", help="No dev options", action="store_false", dest="dev"
    )
    dev_option_group.add_argument(
        "--dev-opt",
        help="add extra dev options",
        nargs="+",
        dest="dev_opts",
        default=[],
    )
    start.add_argument(
        "--load-language",
        help="specifies the languages for the translations you want to be"
        " loaded (also read from configuration file)",
        default=None,
        dest="LOAD_LANGUAGE",
    )
    # TODO there are more
    odoo_log_levels = ["info", "warn", "debug", "debug_sql"]
    # TODO vérifier si fonctionne avec populate
    parser.add_argument(
        "--log-level",
        help="Override odoo log level (for tests and install)",
        default=None,
        choices=odoo_log_levels,
    )
    # TODO vérifier si fonctionne avec populate
    parser.add_argument(
        "--log-handler",
        help="log handler [default: %(default)s]",
        default=None,
        action="append",
    )
    parser.add_argument(
        "--restore",
        help="Database to restore [default: %(default)s]",
        default=None,
        dest="DUMP",
    )
    # TODO vérifier que ça fonctionne correctement sur toutes les commandes
    parser.add_argument(
        "--config",
        help="Configuration file to use, if any [default: %s]"
        % os.path.join("conf", "dev", "odoo.conf"),
    )

    py3o_group = parser.add_mutually_exclusive_group()
    py3o_group.add_argument(
        "--start-py3o",
        help="start py3o docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=1,
        dest="start_py3o",
    )
    py3o_group.add_argument(
        "--no-start-py3o",
        help="do not start py3o docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=-1,
        dest="start_py3o",
    )
    rednerd_group = parser.add_mutually_exclusive_group()
    rednerd_group.add_argument(
        "--start-rednerd",
        help="start rednerd docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=1,
        dest="start_rednerd",
    )
    rednerd_group.add_argument(
        "--no-start-rednerd",
        help="do not start rednerd docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=-1,
        dest="start_rednerd",
    )
    redner_options_group = parser.add_argument_group()
    redner_options_group.add_argument(
        "--redner-db-user",
        help="Database user for redner [default: %(default)s]",
        default="redner",
    )
    redner_options_group.add_argument(
        "--redner-db-password",
        help="Database password for redner [default: %(default)s]",
        default="redner",
    )
    redner_options_group.add_argument(
        "--redner-admin-password",
        help="Redner admin password. Will be updated in redner with this value "
        "[default: %(default)s]",
        default="admin",
    )
    redner_options_group.add_argument(
        "--redner-user",
        help="Redner user that will be used. Created if it does not exists [default: "
        "%(default)s]",
        default="odoo",
    )
    redner_options_group.add_argument(
        "--redner-password",
        help="Redner user that will be used. Will be updated in redner [default: "
        "%(default)s]",
        default="redner-odoo-password",
    )
    redis_group = parser.add_mutually_exclusive_group()
    redis_group.add_argument(
        "--start-redis",
        help="start redis docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=1,
        dest="start_redis",
    )
    redis_group.add_argument(
        "--no-start-redis",
        help="do not start redis docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=-1,
        dest="start_redis",
    )

    parser.add_argument(
        "--odoo-sources",
        help="Odoo sources to use, if any [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--pre-sql-script",
        help="SQL script to run on database [default: %(default)s]",
        default=None,
        dest="PRE_SQL",
    )

    parser.add_argument(
        "--chown",
        help="run chown [default: %(default)s]",
        action="store_true",
        dest="run_chown",
    )

    marabunta_group = start.add_mutually_exclusive_group()
    marabunta_group.add_argument(
        "--no-marabunta",
        help="Do not run marabunta",
        action="store_false",
        default=True,
        dest="marabunta",
    )
    marabunta_options_group = marabunta_group.add_argument_group()
    marabunta_options_group.add_argument(
        "--marabunta-db-user",
        help="specify a db user for marabunta, this is useful if the odoo user is not "
        "able to create databases or extensions",
    )
    marabunta_options_group.add_argument(
        "--marabunta-db-password",
        help="specify a db password for marabunta",
    )
    marabunta_options_group.add_argument(
        "--marabunta-force-version",
        help="marabunta force version argument (see marabunta documentation)",
    )

    pythondevmode_group = parser.add_mutually_exclusive_group()
    pythondevmode_group.add_argument(
        "--pythondevmode",
        help="Set PYTHONDEVMODE",
        action="store_true",
        default=False,
        dest="force_python_dev_mode",
    )
    pythondevmode_group.add_argument(
        "--no-pythondevmode",
        help="Do not set PYTHONDEVMODE",
        action="store_true",
        default=False,
        dest="force_no_python_dev_mode",
    )

    # workers options
    start.add_argument(
        "--workers",
        help="Number of workers [default: %(default)s]",
        default=0,
        type=int,
    )

    parser.add_argument(
        "--activate-s3-storage",
        help="Activate S3 storage using the provided bucket as the main one [default: "
        "%(default)s]",
        default=None,
        type=str,
    )

    parser.add_argument(
        "--queue-job-no-delay",
        help="Set environment variable to run tasks immediately rather than delaying"
        "them",
        action="store_true",
    )

    anthem = command.add_argument_group()
    anthem.add_argument(
        "--anthem-song",
        help="Run an anthem song, not odoo",
        default=None,
    )
    anthem.add_argument(
        "--anthem-path",
        help="Path to a project to install before running a song",
        default=None,
    )

    return parser


def set_rednerd_configparameters(cursor, user, api_key, server_url) -> None:
    for key, value in (
        ("redner.account", user),
        ("redner.api_key", api_key),
        ("redner.server_url", server_url),
    ):
        if value is not None:
            cursor.execute(
                "INSERT INTO ir_config_parameter (key, value) VALUES (%s, %s) ON "
                "CONFLICT (key) DO UPDATE SET value=%s WHERE ir_config_parameter.key = "
                "%s",
                (key, value, value, key),
            )


def main(argv: Optional[List[str]] = None) -> int:
    """Parse arguments and launch conversion"""
    project_path = os.path.realpath(".")
    project_name = os.path.basename(project_path)

    # basic detection to avoid errors
    if project_name == "odoo_scripts":
        logging.fatal(
            "You must run this script from the super project"
            " (./odoo_scripts/docker_dev_start.py)"
        )
        return 1

    # TODO detect that user is member of docker group

    parser = __parser(project_name)

    # TODO add a way to add options to docker
    # TODO add a way to add arg to odoo

    nmspc = parser.parse_args(argv)
    project_name = nmspc.name
    apply(nmspc)
    db_user = nmspc.db_user
    db_password = nmspc.db_password
    use_host_network = nmspc.host_network
    _logger.debug("Use host network: %s", str(use_host_network))
    start_postgresql = nmspc.start_postgresql
    dev = nmspc.dev
    dev_opts = nmspc.dev_opts
    restore_filename = nmspc.DUMP
    database = nmspc.database
    """Odoo database. Optional[str]"""
    load_language = ""
    installing_or_updating = False
    odoo_sources = nmspc.odoo_sources
    pre_sql_script = nmspc.PRE_SQL
    run_chown = nmspc.run_chown
    populate_model = nmspc.populate_model
    populate_size = nmspc.populate_size
    choosen_command: Command
    if nmspc.odoo_help:
        choosen_command = Command.HELP
    elif nmspc.anthem_song:
        choosen_command = Command.ANTHEM
    elif populate_size or populate_model:
        choosen_command = Command.POPULATE
    else:
        choosen_command = Command.RUN
    odoo_help: bool = choosen_command == Command.HELP  # XXX à garder ou pas ?
    marabunta = nmspc.marabunta and choosen_command == Command.RUN
    marabunta_db_user = nmspc.marabunta_db_user
    marabunta_db_password = nmspc.marabunta_db_password
    marabunta_force_version = nmspc.marabunta_force_version
    force_python_dev_mode = nmspc.force_python_dev_mode
    force_no_python_dev_mode = nmspc.force_no_python_dev_mode
    workers = nmspc.workers if choosen_command == Command.RUN else 0
    gevent = workers > 0
    test_tags = nmspc.test_tags
    activate_s3_storage = nmspc.activate_s3_storage
    queue_job_no_delay = nmspc.queue_job_no_delay

    if restore_filename:
        if not database:
            _logger.fatal("No database name given for restore")
            return 13

    config = Config()

    # read from conguration when not set
    if not db_user and config.db_user is not None:
        db_user = config.db_user
    if not db_password and config.db_password is not None:
        db_password = config.db_password
    if not marabunta_db_user and config.marabunta_db_user is not None:
        marabunta_db_user = config.marabunta_db_user
    if not marabunta_db_password and config.marabunta_db_password is not None:
        marabunta_db_password = config.marabunta_db_password
    if config.load_language is not None:
        load_language = config.load_language

    odoo_type = config.odoo_type
    image = config.local_image_name
    postgresql_version = config.postgresql_version
    module_list = config.module_list
    if nmspc.start_py3o:
        _logger.debug("Command line start py3o %d", nmspc.start_py3o)
        start_py3o_stack = nmspc.start_py3o == 1
    else:
        start_py3o_stack = config.start_py3o
        _logger.debug("No command line start py3o %d", start_py3o_stack)

    if nmspc.start_rednerd:
        _logger.debug("Command line start rednerd %d", nmspc.start_rednerd)
        start_rednerd_stack = nmspc.start_rednerd == 1
    else:
        start_rednerd_stack = config.start_rednerd
        _logger.debug("No command line start rednerd %d", start_rednerd_stack)
    # TODO ajouter l’option de démarrer redis dans Configuration
    start_redis: bool = (nmspc.start_redis == 1) if nmspc.start_redis else False

    _logger.debug("Docker image: %s", image)
    # detect if docker image already exists
    if not nmspc.no_build:
        failed = build_local_image(get_build_options(nmspc), nmspc.force_build)
        if failed:
            return failed

    # options is only used with subprocess call
    options: Dict[str, Any] = {
        "name": project_name,
        "add_hosts": [],
        "publish": [],
        "envs": {},
    }
    if activate_s3_storage:
        options["envs"].update(
            {
                "s3AttachmentStorageActivate": "true",
                "s3AttachmentStorageServerBucket": activate_s3_storage,
            }
        )
    if queue_job_no_delay:
        options["envs"]["TEST_QUEUE_JOB_NO_DELAY"] = "1"
    options["hostname"] = project_name
    # TODO detect that from configuration or force it
    exposed_port = 8069
    # TODO detect that from configuration or force it
    odoo_run_port = 8069
    if choosen_command == Command.RUN and not gevent and not use_host_network:
        options["publish"].append((exposed_port, odoo_run_port))
    mounts: List[Mount] = []
    """Mounts for Odoo docker"""
    tested_modules: str = ""
    # TODO handle other version of odoo (base image need to be changed too)
    coverage = (
        (nmspc.test or nmspc.test_default or nmspc.install_test) and not test_tags
    ) and odoo_type not in (ODOO_7, ODOO_8, ODOO_9, ODOO_10)
    arg = []
    if choosen_command == Command.POPULATE:
        arg.append("populate")
    elif choosen_command == Command.RUN:
        arg.extend(
            [
                "coverage-start" if coverage else "start",
                "--workers",
                str(workers),
            ]
        )
    elif choosen_command == Command.HELP:
        arg.append("start")
    elif choosen_command == Command.ANTHEM:
        # TODO fait à la fin de la fonction, à remettre ici
        pass
    if db_password:
        arg.extend(("--db_password", db_password))
    if nmspc.update:
        arg.extend(("-u", nmspc.update, "--i18n-overwrite"))
        installing_or_updating = True
    if nmspc.test or nmspc.test_default or nmspc.install_test:
        arg.append("--test-enable")
    if nmspc.test:
        tested_modules = nmspc.test
        arg.extend(("-u", nmspc.test))
        if odoo_type == ODOO_7:
            arg.append("--log-level=test")
        installing_or_updating = True
    if nmspc.install_test:
        tested_modules = nmspc.install_test
        arg.extend(("-i", nmspc.install_test))
        if odoo_type == "odoo7":
            arg.append("--log-level=test")
        installing_or_updating = True
    if nmspc.test_default:
        test_modules = config.module_list_tests
        tested_modules = ",".join(test_modules)
        arg.extend(("-u", tested_modules))
        if odoo_type == ODOO_7:
            arg.append("--log-level=test")
        arg.append("--stop-after-init")
        installing_or_updating = True
    if tested_modules or nmspc.stop_after_init:
        arg.append("--stop-after-init")
    if database:
        arg.extend(("-d", database))
    if nmspc.install or nmspc.install_default:
        installing_or_updating = True
        modules_to_install = []
        if nmspc.install:
            modules_to_install.extend(nmspc.install.split(","))
        if nmspc.install_default:
            modules_to_install.extend(module_list)
        if modules_to_install:
            arg.append("-i")
            arg.append(",".join(modules_to_install))
    if nmspc.without_demo is not None:
        arg.append("--without-demo")
        arg.append(nmspc.without_demo)
    if nmspc.max_cron_threads:
        arg.append("--max-cron-threads=%s" % nmspc.max_cron_threads)
    if nmspc.log_level:
        arg.append("--log-level")
        arg.append(nmspc.log_level)
    if nmspc.log_handler:
        for lh in nmspc.log_handler + config.log_handler:
            arg.append("--log-handler")
            arg.append(lh)
    if nmspc.LOAD_LANGUAGE:
        load_language = nmspc.LOAD_LANGUAGE
    if load_language and installing_or_updating:
        arg.append("--load-language")
        arg.append(load_language)
    if (populate_model or populate_size) and odoo_type in (
        ODOO_7,
        ODOO_8,
        ODOO_9,
        ODOO_10,
        ODOO_11,
        ODOO_12,
        ODOO_13,
    ):
        _logger.warning("Populate is not available in your version")
    if populate_model:
        arg.append("--model")
        arg.append(populate_model)
    if populate_size:
        arg.append("--size")
        arg.append(populate_size)

    # auto detect local ip
    if use_host_network:
        local_ip = "127.0.0.1"
        options["networks"] = ["host"]
    else:
        local_ip = None
        try:
            from netifaces import AF_INET, ifaddresses  # type: ignore[import]

            iface_name = "docker0"
            addresses = [
                i["addr"]
                for i in ifaddresses(iface_name).setdefault(
                    AF_INET, [{"addr": "No IP addr"}]
                )
            ]
            if addresses:
                local_ip = addresses[0]
        except ImportError:
            _logger.warning(
                "Consider installing python netifaces to ease local IP detection"
            )
        if not local_ip:
            _logger.info("Contacting Google Public DNS to find our IP")
            for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]:
                s.connect(("8.8.8.8", 53))
                local_ip = s.getsockname()[0]
                s.close()
                break

        _logger.debug("IP found %s", local_ip)

    # auto detect local conf
    user = db_user
    password = db_password
    local_conf_path = None
    cp_local = None
    if nmspc.config:
        local_conf_path = nmspc.config
        if os.path.isfile(local_conf_path):
            _logger.info("Local Odoo configuration file found: %s", local_conf_path)
            cp_local = ConfigParser()
            cp_local.read(local_conf_path)
        else:
            _logger.critical("No configuration file at: %s", local_conf_path)
            return 82
    else:
        local_conf_path = os.path.join("conf", "dev", "odoo.conf")
        if os.path.isfile(local_conf_path):
            _logger.info(
                "Default local Odoo configuration file found: %s", local_conf_path
            )
            cp_local = ConfigParser()
            cp_local.read(local_conf_path)
        else:
            _logger.debug(
                "No default local Odoo configuration file found (%s)", local_conf_path
            )
    if local_conf_path and cp_local:
        if odoo_type in (ODOO_7, ODOO_8, ODOO_9, ODOO_10):
            odoo_conf_file = "/opt/odoo/etc/odoo.conf"
        else:
            odoo_conf_file = "/etc/opt/odoo/odoo.conf"
            # change ODOO_RC for version<ODOO_16
            options["envs"]["ODOO_RC"] = odoo_conf_file
        mounts.append(
            mount(
                odoo_conf_file,
                os.path.join(project_path, local_conf_path),
                "bind",
            )
        )
        if not user:
            user = cp_local.get("options", "db_user", fallback=None)
            _logger.debug(
                "Using db user %s as set in Odoo configuration file %s",
                user,
                local_conf_path,
            )
        if not password:
            password = cp_local.get("options", "db_password", fallback=None)
        # doing this avoid having to change the value of redis_host
        # used by smile_redis_session_store
        redis_host = (
            cp_local.has_section("options")
            and cp_local.has_option("options", "redis_host")
            and cp_local.get("options", "redis_host")
        ) or None
        if redis_host:
            options["add_hosts"].append((redis_host, local_ip))
    # default values if nothing else
    if not user:
        # Odoo uses default if nothing is set (including in PGUSER env var)
        if start_postgresql:
            _logger.info("No user defined, using odoo when starting postgres")
            user = "odoo"
        else:
            # use environment variable
            user = os.getenv("PGUSER")
            # defaulting to odoo
            if user:
                _logger.info("Using PGUSER for database user")
            else:
                user = "odoo"
                _logger.info("Defaulting to odoo for database user")
    arg.append("--db_user")
    arg.append(user)

    # directories to run chown on
    chown_directories = set()

    # data volume handling
    base_data_volume_name = "{}_data".format(project_name)
    if odoo_type == ODOO_7:
        # this path is in odoo code, the last part is the system user
        session_volume_name = "{}{}".format(base_data_volume_name, "_sessions")
        target = "/tmp/oe-sessions-{}".format("odoo")
        docker.volume.create(session_volume_name, labels={"mount-location": target})
        mounts.append(mount(target, session_volume_name))
        # by default odoo store attachments in database but it can be set on
        # disk. mount the volume in all cases.
        # XXX data is the path set in the database (in the form file:///data)
        #  it should be read from there, in that case target directory needs
        #  to include the database name.
        #  our databases only use file:///data so this should be fine
        target = (
            "/usr/local/lib/python2.7/dist-packages/"
            "openerp-7.0-py2.7.egg/openerp/data"
        )
        filestore_volume_name = "{}_filestore".format(base_data_volume_name)
        docker.volume.create(filestore_volume_name, labels={"mount-location": target})
        mounts.append(mount(target, filestore_volume_name))
    else:
        data_volume_path = "/var/lib/data"
        mounts.append(mount(data_volume_path, base_data_volume_name))
        _, created = DockerClient.create_volume(
            base_data_volume_name,
            {
                "mount-location": data_volume_path,
            },
        )
        if created:
            chown_directories.add(data_volume_path)
        arg.append("--data-dir")
        arg.append(data_volume_path)
        if run_chown:
            chown_directories.add(data_volume_path)

    # make sure the permission in the volumes are correct
    if run_chown:
        chown_directories.add("/mnt")

    if chown_directories:
        _logger.info("chown container’s %s to Odoo user", ", ".join(chown_directories))
        docker.run(
            image,
            command=["odoo", "--recursive", *list(chown_directories)],
            remove=True,
            entrypoint="/bin/chown",
            mounts=mounts,
            user="root",
        )

    if start_py3o_stack:
        start_py3o()
        # provide a couple of handy dns name for py3o
        options["add_hosts"].append(("py3o", local_ip))
        options["add_hosts"].append(("py3o-fusion-server", local_ip))

    # Database connection

    dbport = None
    if nmspc.dbport:
        dbport = str(nmspc.dbport)
        arg.append("--db_port")
        arg.append(dbport)

    db_host = nmspc.db_host

    create_database = False
    # Check that connection can be done, try to create user if asked to
    # TODO: handle the case where the database is still starting up
    # TODO: (and remove the sleep)
    if not odoo_help:
        if start_postgresql:
            if db_host:
                _logger.warning("Ignoring given db_host when starting postgres")
            if dbport:
                _logger.warning("Ignoring given db_port when starting postgres")
            db_host = "/var/run/postgresql"
            dbport = "5432"
            # Use socket rather than export port (avoid having to find a
            # free port number)
            pg, _stop_postgresql, socket_path = docker_run_postgresql(
                project_name, postgresql_version, host_pg_port=None
            )
            if socket_path is not None:
                mounts.append(mount(db_host, socket_path, "bind"))
        _logger.info("Testing database connection on base postgres")
        try:
            if start_postgresql:
                connect(
                    user=user,
                    password=password,
                    database="postgres",
                    host=socket_path,
                    port=dbport,
                )
                _logger.info("DB connection successful")
            else:
                # psql uses PGHOST if provided and a socket otherwise
                if db_host:
                    pass
                elif "PGHOST" in os.environ:
                    db_host = os.getenv("PGHOST")
                    _logger.debug("Using PGHOST=%s", db_host)
                else:
                    # use socket, trying to guess it
                    # db_host = "/var/run/postgresql"
                    # _logger.debug("Defaulting to host %s", db_host)
                    db_host = local_ip
                if db_host.startswith("/"):
                    # mount the path inside the docker
                    mounts.append(mount(db_host, db_host, "bind"))
                connect(
                    user=user,
                    password=password,
                    database="postgres",
                    host=db_host,
                    port=dbport,
                )
                _logger.info("DB connection successful")

        except OperationalError as exception:
            if nmspc.create_user:
                _logger.debug("Cannot connect to database with user %s", user)
                _logger.debug(exception)
                _logger.info("Creating user %s", user)
                if start_postgresql:
                    connection = connect(
                        user="postgres",
                        database="postgres",
                        host=socket_path,
                        port=5432,
                        password=POSTGRES_PASSWORD,
                    )
                else:
                    loginname = pwd.getpwuid(os.getuid())[0]
                    # use socket to create user
                    # TODO test if removing more arguments works
                    connection = connect(user=loginname, database="postgres")
                _logger.info("User created")
                with connection.cursor() as cursor:
                    # not injection safe but you are on your own machine
                    # with already full access to db
                    cursor.execute(
                        "CREATE ROLE %s LOGIN CREATEDB PASSWORD %%s" % user,
                        (password,),
                    )
                connection.commit()
                connection.close()
            else:
                _logger.critical("Cannot connect to database with user %s", user)
                _logger.critical(exception)
                _logger.info(
                    "You can add the --create-user argument to try to create it"
                )
                return 16

        arg.append("--db_host")
        arg.append(db_host)

        # Create database if it does not exists

        if database:
            with connect(
                user=user,
                password=password,
                database="postgres",
                host=socket_path if start_postgresql else db_host,
                port=dbport,
            ) as connection:
                with connection.cursor() as cursor:
                    cursor.execute(
                        "SELECT count(*) FROM pg_database WHERE datname=%s", (database,)
                    )
                    data = cursor.fetchall()
                    create_database = data[0][0] == 0

    # --- restore / create database ---
    # TODO find out why odoo_help would stop us from restoring
    if start_postgresql and not odoo_help and (restore_filename or create_database):
        _logger.info("Creating database %s for %s", database, user)
        pg.execute(["createdb", "-U", user, database])
    if start_postgresql and not odoo_help and restore_filename:
        restore_basename = os.path.basename(restore_filename)
        _logger.info("Copying dump file in docker")
        inside_restore_filename = "/tmp/{}".format(restore_basename)
        docker.copy(restore_filename, (pg, inside_restore_filename))
        _logger.info("Restoring database %s", database)
        with CaptureDockerExitCode() as exit_code:
            pg.execute(
                [
                    "pg_restore",
                    "-U",
                    user,
                    "--no-owner",
                    "--no-acl",
                    "--job=4",
                    "--dbname",
                    database,
                    inside_restore_filename,
                ]
            )
        if int(exit_code):
            _logger.fatal("Error when restoring database")
            return 15
        _logger.info("Removing dump file in docker")
        pg.execute(["rm", inside_restore_filename])

    if not start_postgresql and not odoo_help and (restore_filename or create_database):
        connection = connect(
            user=user,
            password=password,
            database="postgres",
            host=db_host,
            port=dbport,
        )
        connection.autocommit = True
        _logger.info("Creating database %s", database)
        with connection.cursor() as cursor:
            cursor.execute(f'CREATE DATABASE "{database}"')
        connection.close()
    if not start_postgresql and not odoo_help and restore_filename:
        _logger.info("Restoring database %s", database)
        env = {
            "PGHOST": db_host,
            "PGPASSWORD": password,
            "PGUSER": user,
            "PG_COLOR": "always",
        }
        if dbport is not None:
            env["PGPORT"] = dbport
        restore = call(
            [
                "pg_restore",
                "--no-owner",
                "--no-acl",
                "--job=4",
                "--dbname",
                database,
                restore_filename,
            ],
            env=env,
        )
        if restore != 0:
            _logger.fatal("Errors during restoration")
            return 15

    # --- Pre SQL script ---
    if pre_sql_script:
        if start_postgresql:
            pre_sql_script_basename = os.path.basename(pre_sql_script)
            _logger.info("Copying SQL script %s in docker", pre_sql_script_basename)
            inside_pre_sql_script = "/tmp/{}".format(pre_sql_script_basename)
            docker.copy(pre_sql_script, (pg, inside_pre_sql_script))
            _logger.info(
                "Running SQL script %s in %s",
                pre_sql_script_basename,
                database,
            )
            with CaptureDockerExitCode() as exit_code:
                pg.execute(
                    [
                        "psql",
                        "-U",
                        user,
                        "-d",
                        database,
                        "-f",
                        inside_pre_sql_script,
                    ]
                )
            if int(exit_code) > 0:
                _logger.fatal("Error when running SQL script %s", pre_sql_script)
                return 16
            _logger.info("Removing SQL script %s in docker", pre_sql_script)
            # remove script
            pg.execute(["rm", inside_pre_sql_script])
        else:
            _logger.info("Running SQL script %s on %s", pre_sql_script, database)
            env = {
                "PGHOST": db_host,
                "PGPASSWORD": password,
                "PGUSER": user,
                "PG_COLOR": "always",
            }
            if dbport is not None:
                env["PGPORT"] = dbport
            script = call(["psql", "--dbname", database, "-f", pre_sql_script], env=env)
            if script > 0:
                _logger.fatal("Error when running SQL script %s", script)
                return 16

    # --- Rednerd start ---

    if start_rednerd_stack:
        redner_db_user = nmspc.redner_db_user
        redner_db_password = nmspc.redner_db_password
        redner_admin_password = nmspc.redner_admin_password
        redner_user = nmspc.redner_user
        redner_password = nmspc.redner_password
        # share redner database by project rather than by database
        redner_database = get_default_redner_database(project_name)
        connect_kwargs: Dict[str, Any] = {
            "database": "postgres",
        }
        if start_postgresql:
            # uses the unix socket
            connect_kwargs.update(
                {
                    "user": "postgres",
                    "host": db_host,
                    "port": "5432",
                    "password": POSTGRES_PASSWORD,
                }
            )
        else:
            loginname = pwd.getpwuid(os.getuid())[0]
            # use socket to create user
            connect_kwargs.update({"user": loginname})
        connection = connect(**connect_kwargs)
        connection.autocommit = True
        with connection.cursor() as cursor:
            if nmspc.create_user:
                try:
                    cursor.execute(
                        "CREATE ROLE %s LOGIN PASSWORD %%s" % redner_db_user,
                        (redner_db_password,),
                    )
                    _logger.info("Created database role %s", redner_db_user)
                except DuplicateObject:
                    _logger.debug("Database role exists")
            try:
                _logger.info("Creating database %s", redner_database)
                cursor.execute(f'CREATE DATABASE "{redner_database}"')
                _logger.info("Database created")
            except DuplicateDatabase:
                # ignore exception if the database already exists
                _logger.info("Database already exists")
        connect_kwargs["database"] = redner_database
        # do that here as we have superuser rights, while redner user probably does not
        with connect(**connect_kwargs) as connection:
            with connection.cursor() as cursor:
                _logger.info("Creating extension hstore if necessary")
                cursor.execute("CREATE EXTENSION IF NOT EXISTS hstore")
        if config.rednerd_repository is None:
            _logger.fatal(
                "Missing rednerd image name (rednerd_repository) in configuration"
            )
            return 81
        redner_docker = DockerRednerd(
            project_name,
            config.rednerd_repository,
            db_user=redner_db_user,
            db_password=redner_db_password,
            database=redner_database,
            db_host=db_host,
            db_port=dbport,
            attach=False,
            redner_user=redner_user,
            redner_password=redner_password,
            redner_admin_password=redner_admin_password,
            use_socket=True,
            version=config.rednerd_version,
        )
        redner_docker.start()
        # only useful when using port rather than socket. Not used at the moment
        options["add_hosts"].append(("redner", local_ip))
        # bind the socket dir in Odoo
        if hasattr(redner_docker, "socket_dir"):
            mounts.append(
                mount(redner_docker.socket_dir, redner_docker.socket_dir, "bind")
            )

        # --- Rednerd change config parameters in Odoo database ---
        if database:
            server_url = None
            # use unix socket if there is one, port otherwise.
            if hasattr(redner_docker, "socket_path"):
                server_url = redner_docker.socket_path
            elif hasattr(redner_docker, "redner_port"):
                server_url = f"http://localhost:{redner_docker.redner_port}/api/"
            _logger.info(
                "Changing redner ir.config_parameters in database %s", database
            )
            # connect to odoo database and set the ir.config_parameters
            with connect(
                user=user,
                password=password,
                database=database,
                host=db_host,
                port=dbport,
            ) as connection:
                with connection.cursor() as cursor:
                    set_rednerd_configparameters(
                        cursor, redner_user, redner_docker.api_key, server_url
                    )

    # --- Redis start ---

    redis_docker: Optional[DockerRedis] = None
    if start_redis:
        # TODO stop at exit à vrai sinon s’éteint pas?
        redis_docker = DockerRedis(project_name, attach=False)
        redis_docker.start()
        # bind the socket dir in Odoo
        mounts.append(mount(redis_docker.socket_dir, redis_docker.socket_dir, "bind"))
        options["envs"].update(
            # Variables used by flanker.
            REDIS_PORT="0",
            REDIS_DB="0",
            # Rely on a flanker patch to work.
            REDIS_UNIX_SOCKET_PATH=redis_docker.socket_path,
            # Variables used by session_redis
            ODOO_SESSION_REDIS="1",
            ODOO_SESSION_REDIS_URL=f"unix://{redis_docker.socket_path}?db=1",
        )

    # volume magic
    mount_list, _ = modules_mount(config)
    if mount_list:
        mounts.extend(mount_list)
    if odoo_type in (ODOO_7, ODOO_8):
        arg.append("--addons-path=/opt/odoo/sources/odoo/addons")
    # No way to set addons path to nothing for other versions unfortunately

    apply_flake8(nmspc, config)
    apply_isort(nmspc, config)
    apply_pylint(nmspc, config)

    # developer mode options
    if choosen_command == Command.RUN and dev:
        if odoo_type not in (ODOO_7, ODOO_8, ODOO_9):
            # automatically add reload if not already asks for
            if not any(opt in dev_opts for opt in ("all", "reload")):
                dev_opts.append("reload")
            # XXX pdb should not be run if not interactive
            # add pdb if no pubd, wdb, ipdb, pdb or all
            # the list of available options differs by odoo version and can be
            # seen with odoo --help, available with --odoo-help
            if not any(
                debugger in dev_opts
                for debugger in ("pubd", "wdb", "ipdb", "pdb", "all")
            ):
                dev_opts.append("pdb")
            # Add werkzeug if not already enabled
            if odoo_type not in (ODOO_10, ODOO_11, ODOO_12, ODOO_13, ODOO_14):
                if not any(opt in dev_opts for opt in ("all", "werkzeug")):
                    dev_opts.append("werkzeug")
            arg.append(f"--dev={','.join(set(dev_opts))}")
        if odoo_type in (ODOO_8,):
            arg.append("--auto-reload")
    if force_python_dev_mode or (
        config.pythondevmode and not force_no_python_dev_mode and not odoo_help
    ):
        options["envs"]["PYTHONDEVMODE"] = 1

    if choosen_command == Command.RUN and test_tags:
        arg.extend(("--test-tags", test_tags))

    sentry_release = None
    project_version = "local"

    package = None
    if os.path.exists("pyproject.toml"):
        pyproject_data = None
        with open("pyproject.toml", "rb") as f:
            pyproject_data = load(f)
        if (
            pyproject_data
            and "project" in pyproject_data
            and "name" in pyproject_data["project"]
        ):
            package = pyproject_data["project"]["name"]

    if os.path.exists(".hg"):
        # Next version guessing could be done by using marabunta yaml and using it
        # rather than doing latesttag.1
        mercurial = hglib.open(project_path)
        project_version = mercurial.rawcommand(
            hglib.client.cmdbuilder(
                b"identify",
                template=b"{latesttag}{if('{dirty}', '.1.dev{latesttagdistance}{if('{to"
                b"pic}', ' {fqbn}')}', '{ifeq('{latesttagdistance}', 1, '', '{ifeq('{la"
                b"testtagdistance}', 2, '', '.1.dev{latesttagdistance}{if('{topic}', ' "
                b"{fqbn}')}')}')}')}",
            )
        ).decode("UTF-8")
        # do not include fqbn for sentry, sentry will not recognize a version name
        # containing a slash
        sentry_release = mercurial.rawcommand(
            hglib.client.cmdbuilder(
                b"identify",
                template=b"{latesttag}{if('{dirty}', '.1.dev{latesttagdistance}', '{ife"
                b"q('{latesttagdistance}', 1, '', '{ifeq('{latesttagdistance}', 2, '', "
                b"'.1.dev{latesttagdistance}')}')}')}",
            )
        ).decode("UTF-8")

    if os.path.exists(".git"):
        # TODO use gitpython
        sentry_release = project_version = check_output(["git", "describe"]).decode()

    options["envs"]["VERSION_PACKAGE"] = ""
    options["envs"]["SENTRY_RELEASE"] = (
        f"{package}@{sentry_release}" if package else sentry_release
    )
    options["envs"]["VERSION"] = project_version

    if odoo_help:
        # Plenty of extra arguments but it works
        arg.append("--help")

    # bind Odoo sources
    if odoo_sources:
        _logger.info("Binding Odoo sources at %s", odoo_sources)
        mounts.extend(odoo_source_mounts(config, odoo_sources))

    # avoid chowning at each run, faster but can cause trouble
    # newer odoo images does not use this environment variable anymore
    options["envs"]["ODOO_CHOWN"] = "false"

    options["envs"]["PGSSLMODE"] = "disable"

    if coverage:
        module_prefix = "odoo.addons."
        image_env = docker.image.inspect(image).config.env
        if image_env is not None:
            for env_line in image_env:
                key, value = env_line.split("=", maxsplit=1)
                if key == "ODOO_BASE_PATH":
                    module_prefix = value + "/addons/"
                    break
        source_files = ",".join(
            [f"{module_prefix}{module}" for module in tested_modules.split(",")]
        )
        options["envs"]["COVERAGE_RUN_OPTIONS"] = (
            "--source=" + source_files + " --omit=*/__manifest__.py --branch"
        )

    if marabunta:
        marabunta_migration_file = config.marabunta_migration_file
        # if nothing is set, set the value to marabunta.yaml if there is such a file
        if not marabunta_migration_file:
            if os.path.exists("marabunta.yaml"):
                marabunta_migration_file = "marabunta.yaml"
    else:
        marabunta_migration_file = None

    if marabunta_migration_file is not None:
        marabunta_container_path = "/etc/marabunta.yaml"
        # add marabunta file to mounts
        mounts.append(
            mount(
                marabunta_container_path,
                os.path.join(project_path, marabunta_migration_file),
                "bind",
                True,
            )
        )
        mounts.extend(anthem_mounts(config))

        options["envs"].update(
            {
                "MARABUNTA_MIGRATION_FILE": marabunta_container_path,
                "MARABUNTA_MODE": "dev",
                "MARABUNTA_ALLOW_SERIE": "on",
            }
        )
        if start_postgresql:
            # TODO à tester, je pense que postgres ne se connecte pas
            options["envs"]["MARABUNTA_DB_USER"] = marabunta_db_user or "postgres"
        # marabunta is run inside the container by the start command
    else:
        # Only set to avoid problems with image that inherit another image using
        # marabunta while not using marabunta themselves
        options["envs"]["MARABUNTA_MIGRATION_FILE"] = ""
    if marabunta_force_version:
        options["envs"]["MARABUNTA_FORCE_VERSION"] = marabunta_force_version
    if marabunta_db_user:
        # only set if provided a value, otherwise the image will get the value from
        # command line arguments or the configuration file (using the same user os
        # the one used in odoo)
        options["envs"]["MARABUNTA_DB_USER"] = marabunta_db_user
    if marabunta_db_password:
        options["envs"]["MARABUNTA_DB_PASSWORD"] = marabunta_db_password

    if gevent:
        gevent_port = find_free_port()
        run_kwargs: Dict[str, Any] = {
            "command": ["caddy", "run", "--config", "/etc/caddy/Caddyfile"],
            "name": f"{project_name}_caddy",
            "mounts": [],
        }
        # force the port, so that the caddyfile is correct whatever the configuration
        if use_host_network:
            caddy_port = str(odoo_run_port)
            # change the port odoo runs on
            # TODO hopefully does not return 8069/odoo_run_port
            odoo_run_port = find_free_port()
            options["envs"]["MARABUNTA_WEB_PORT"] = str(odoo_run_port)
        else:
            caddy_port = "2019"
        arg.append("--proxy-mode")
        # TODO pre odoo 13: check if http-port should be changed to xmlrpc-port
        gevent_uri = "/websocket"
        gevent_option = "--gevent-port"
        if odoo_type in (
            ODOO_7,
            ODOO_8,
            ODOO_9,
            ODOO_10,
            ODOO_11,
            ODOO_12,
            ODOO_13,
            ODOO_14,
            ODOO_15,
        ):
            gevent_uri = "/longpolling"
            gevent_option = "--longpolling-port"
        arg.extend(
            (
                gevent_option,
                str(gevent_port),
                "--http-port",
                str(odoo_run_port),
            )
        )

        odoo_host = "" if use_host_network else project_name
        caddyfile = f"""{{
    log {{
        format console
    }}
}}

:{caddy_port} {{
    reverse_proxy {gevent_uri}/* {odoo_host}:{gevent_port}
    reverse_proxy {odoo_host}:{odoo_run_port}
}}
"""
        with tempfile.NamedTemporaryFile(
            delete=False, prefix="Caddyfile."
        ) as caddyfile_tmp:
            caddyfile_tmp.write(caddyfile.encode("UTF-8"))

        def cleanup_caddyfile():
            os.unlink(caddyfile_tmp.name)

        atexit.register(cleanup_caddyfile)

        run_kwargs["mounts"].append(
            mount("/etc/caddy/Caddyfile", caddyfile_tmp.name, "bind")
        )
        if use_host_network:
            run_kwargs["networks"] = ["host"]
        else:
            docker.network.create(name=project_name, driver="bridge")
            run_kwargs.update(
                {
                    "network": project_name,
                    "ports": {caddy_port: odoo_run_port},
                    "hostname": "caddy",
                }
            )
            options["network"] = project_name

        _logger.info(
            "Starting caddy, use http://localhost:%s/ to access Odoo", caddy_port
        )
        # TODO there is an error as Odoo is not started up yet when this is started
        #  so maybe start it after Odoo started? (only happens if odoo windows are
        #  opened)
        # TODO also error when Odoo is powered off
        DockerClient.launch("caddy", "latest", run_kwargs)

    if choosen_command == Command.ANTHEM:
        options["name"] += "-anthem"
        if nmspc.anthem_path:
            mounts.extend(
                [
                    mount(
                        "/tmp/anthem_song",
                        os.path.realpath(nmspc.anthem_path),
                        "bind",
                        readonly=True,
                    ),
                    mount("/var/cache/pip", "odoo_scripts-pip-cache"),
                ]
            )
            options["entrypoint"] = "zsh"
            options["envs"]["PIP_CACHE_DIR"] = "/var/cache/pip"
            # TODO error code not returned
            # all the arguments are passed at the end of anthem arguments (see its help)
            arg = [
                "-c",
                "pip install /tmp/anthem_song;exec anthem "
                + nmspc.anthem_song
                + " "
                + " ".join(arg),
            ]
        else:
            options["entrypoint"] = "anthem"
            # all the arguments are passed at the end of anthem arguments (see its help)
            arg = [nmspc.anthem_song, *arg]

    _logger.info("Starting Odoo")
    options["envs"]["scriptLoggingLevel"] = "INFO"
    with CaptureDockerExitCode() as exit_code:
        docker.container.run(
            image,
            command=arg,
            mounts=mounts,
            **options,
            tty=True,
            interactive=True,
            remove=True,
        )
    # Stop redis (python does not exit as the log process is still running)
    if redis_docker:
        _logger.info("Stopping redis")
        redis_docker.stop()
    return int(exit_code)


def find_free_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))
