#!/usr/bin/env python3
"""Run flake8 in a docker"""

import argparse
import logging
import os
import sys
from typing import List, Optional, Tuple

from ._config import (
    ODOO_7,
    ODOO_8,
    ODOO_9,
    ODOO_10,
    ODOO_11,
    ODOO_12,
    ODOO_13,
    ODOO_14,
    ODOO_15,
    ODOO_16,
    Config,
    Configuration,
)
from ._docker_client import DockerClient
from ._parsing import apply, basic_parser

__version__ = "1.5.0"
__date__ = "2020-09-10"
__updated__ = "2024-04-05"

_logger = logging.getLogger(__name__)


def __parser():
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_short_description = __doc__.split(".", maxsplit=1)[0]
    program_license = f"""{program_short_description}

      Created by Vincent Hatakeyama on {__date__}.
      Copyright 2020, 2022-2024 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """
    parser = basic_parser(program_license, program_version_message)
    return parser


def parser_add_flake8_group(parser: argparse.ArgumentParser):
    """Add flake8 option (--flake8/--no-flake8) to the given parser.
    Defaults to not running flake8."""
    flake8_group = parser.add_mutually_exclusive_group()
    flake8_group.add_argument(
        "--flake8",
        help="Run flake8",
        action="store_true",
        dest="flake8",
        default=False,
    )
    flake8_group.add_argument(
        "--no-flake8",
        help="Do not run flake8 [default]",
        action="store_false",
        dest="flake8",
        default=False,
    )


def apply_flake8(namespace: argparse.Namespace, config: Configuration):
    """Run flake8 if the option was set."""
    if namespace.flake8:
        if uses_flake8(config):
            flake8(config)
        else:
            _logger.warning("No flake8 for %s, ignoring running it", config.odoo_type)


def main(argv: Optional[List[str]] = None) -> int:
    """Copy modules for a build, callable version that parses arguments"""
    parser = __parser()
    nmspc = parser.parse_args(argv)
    apply(nmspc)
    config = Config()
    return flake8(config)


def flake8_image(config: Configuration) -> Optional[Tuple[str, str]]:
    """determine image to use based on odoo version"""
    if config.odoo_type in (ODOO_7, ODOO_8, ODOO_9, ODOO_10):
        return "xcgd/flake8", "2"
    if config.odoo_type in (ODOO_14, ODOO_15, ODOO_16):
        return config.local_image, config.local_tag
    # TODO vérifier ce qu’il faut faire en v11 et 13
    if config.odoo_type in (ODOO_11, ODOO_12, ODOO_13):
        return "xcgd/flake8", "3"
    return None


def uses_flake8(config: Configuration) -> bool:
    """Return true if flake8 is used for the modules of the odoo type defined in
    the configuration."""
    return flake8_image(config) is not None


def flake8(config: Configuration, pull: bool = True) -> int:
    """Run flake8"""
    result = flake8_image(config)
    if not result:
        raise Exception(f"No flake8 for {config.odoo_type}")
    repository, tag = result
    _logger.info("Running flake8")
    pwd = os.environ["PWD"]
    return DockerClient.run(
        repository,
        tag,
        {
            "entrypoint": "flake8",
            "command": ["/mnt"],
            "volumes": {pwd: {"bind": "/mnt", "mode": "ro"}},
            "workdir": "/mnt",
        },
        pull,
        {"/mnt": pwd},
    )


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))
