#!/usr/bin/env python3
"""Method to inject environment variables into an Odoo configuration file."""

import logging
import os.path
import sys
from configparser import ConfigParser
from typing import Collection, Optional, Sequence, Tuple

from ._parsing import apply, basic_parser

__version__ = "1.0.1"
__date__ = "2022-02-22"
__updated__ = "2022-12-16"

# tuple with environment variable nome, section and option
_environment_variables: Collection[Tuple[str, str, str]] = (
    ("ODOO_OPTIONS_DB_USER", "options", "db_user"),
    ("ODOO_OPTIONS_DB_PASSWORD", "options", "db_password"),
    ("ODOO_OPTIONS_DB_NAME", "options", "db_name"),
    ("ODOO_OPTIONS_DB_HOST", "options", "db_host"),
    ("ODOO_OPTIONS_DATA_DIR", "options", "data_dir"),
    ("ODOO_OPTIONS_MAX_CRON_THREADS", "options", "max_cron_threads"),
    ("ODOO_OPTIONS_LOAD_LANGUAGE", "options", "load_language"),
    ("ODOO_OPTIONS_ADDONS_PATH", "options", "addons_path"),
)

_logger = logging.getLogger(__name__)


class TargetExistsException(Exception):
    """Exception when the target already exists"""


def inject_env_var(source_path: str, target_path: str, overwrite: bool = False):
    """Inject environment variables into an Odoo configuration file.
    :raises TargetExistsException: when the target already exists
    """
    if os.path.exists(target_path) and not overwrite:
        raise TargetExistsException(f"{target_path} already exists")
    config_parser = ConfigParser()
    config_parser.read(source_path)
    with open(target_path, "w", encoding="UTF-8") as file:
        for env_var_name, section, option in _environment_variables:
            if env_var_name in os.environ:
                _logger.debug("Using %s", env_var_name)
                if not config_parser.has_section(section):
                    config_parser.add_section(section)
                config_parser.set(section, option, os.environ[env_var_name])
        config_parser.write(file)


def main(args: Optional[Sequence[str]] = None) -> int:
    """Parse arguments and launch injection"""
    if args is None:
        args = []

    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_short_description = __doc__.split(".", maxsplit=1)[0]
    program_license = f"""{program_short_description}

  Created by Vincent Hatakeyama on {__date__}.
  Copyright 2022 XCG Consulting. All rights reserved.

  Licensed under the MIT License

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
"""
    # Argument parsing
    parser = basic_parser(program_license, program_version_message)
    parser.add_argument(
        "--source",
        help="source file",
        required=True,
    )
    parser.add_argument(
        "--target",
        help="target file",
        required=True,
    )
    parser.add_argument(
        "--overwrite",
        help="Overwrite target file [default: %(default)s]",
        action="store_true",
        default=False,
    )

    namespace = parser.parse_args(args)
    apply(namespace)
    inject_env_var(namespace.source, namespace.target, namespace.overwrite)
    return 0


if __name__ == "__main__":
    sys.exit(main())
