#!/usr/bin/env python3
"""This script update another repository that is a copy of the sources"""

import argparse
import logging
import os
import shutil
import sys
import tempfile
from subprocess import call
from typing import List, Optional

from ._config import Config, Configuration
from ._cvs import Repo, clone_from, get_repo, repo_kind
from ._parsing import apply, basic_parser
from .docker_build_copy import _find_cvs_parent

_logger = logging.getLogger(__name__)

__version__ = "1.4.0"
__date__ = "2020-06-19"
__updated__ = "2024-11-06"


def __parser():
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = f"""{program_shortdesc}

      Created by Vincent Hatakeyama on {__date__}.
      Copyright 2018, 2020, 2021, 2023, 2024 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """
    parser = basic_parser(program_license, program_version_message)
    add_parser_options(parser)
    return parser


def add_parser_options(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-t",
        "--target",
        help="Target directory. Defaults to a temporary directory",
        default=None,
    )
    parser.add_argument(
        "--no-clean-up",
        help="Do not delete target directory at the end.",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--no-push",
        help="Do not push.",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--apply-tags",
        help="Apply current tags to duplicate",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--force-tags",
        help="Force tags on duplicate",
        default=False,
        action="store_true",
    )
    command = parser.add_mutually_exclusive_group()
    command.add_argument(
        "--copy-only",
        default=False,
        action="store_true",
    )
    cvs = command.add_argument_group()
    cvs.add_argument(
        "--kind",
        help="CVS kind of the sources duplicate. Defaults to same as source.",
        default=None,
    )
    cvs.add_argument(
        "--url",
        help=(
            "CVS URL of the sources duplicate. Defaults to appending _sources to "
            "default path"
        ),
        default=None,
    )
    cvs.add_argument(
        "-b",
        "--branch",
        help="CVS branch to update on the sources duplicate",
        default=None,
    )
    cvs.add_argument(
        "-m",
        "--message",
        help="Commit message",
        default="Update sources",
    )


class Options:
    def __init__(
        self,
        target: Optional[str],
        clean_up: bool,
        copy_only: bool,
        cvs_kind: Optional[repo_kind],
        cvs_url: Optional[str],
        cvs_branch: Optional[str],
        message: str,
        push: bool,
        apply_tags: bool,
        force_tags: bool,
    ):
        self.source_directory = "."
        self.target = target
        self.clean_up = clean_up
        self.copy_only = copy_only
        self.cvs_kind = cvs_kind
        self.cvs_url = cvs_url
        self.cvs_branch = cvs_branch
        self.message = message
        self.push = push
        self.apply_tags = apply_tags
        self.force_tags = force_tags


def get_options(namespace: argparse.Namespace) -> Options:
    return Options(
        namespace.target,
        not namespace.no_clean_up,
        namespace.copy_only,
        namespace.kind,
        namespace.url,
        namespace.branch,
        namespace.message,
        not namespace.no_push,
        namespace.apply_tags,
        namespace.force_tags,
    )


def main(argv: Optional[List[str]] = None) -> int:
    """Copy modules for a build, callable version that parses arguments"""
    parser = __parser()
    nmspc = parser.parse_args(argv)
    apply(nmspc)
    options = get_options(nmspc)
    _update_duplicate_sources(options)
    return 0


def _update_duplicate_sources(options: Options, path: Optional[str] = None):
    """
    :arg directory: path to the base directory
    :arg path: path to use (from `.hg/hgrc`)
    """
    repo = get_repo(options.source_directory)
    # get branch and tags
    if options.cvs_branch is None:
        branch = repo.current_branch
        _logger.info("Current branch %s", branch)
    else:
        branch = options.cvs_branch
    tags = repo.current_tags
    if tags:
        _logger.info("Current tags: %s.", ":".join(tags))
    else:
        _logger.info("No tag")
    # find the duplicate repository URL
    conf = Config()
    duplicate_url: str
    if options.cvs_url:
        duplicate_url = options.cvs_url
    elif conf.duplicate_repo:
        # TODO voir si toujours utile de mettre dans le setup.cfg
        duplicate_url = conf.duplicate_repo
    else:
        # TODO est-ce une bonne idée d’avoir ce truc par défaut ? utilisé sur LEP/JAL
        remote_path = repo.remote_url(path)
        duplicate_url = "{}_sources".format(remote_path)
    _logger.info("Duplicate URL %s", duplicate_url)
    # clone the duplicate repository
    duplicate_destination: str
    if options.target is None:
        duplicate_destination = tempfile.mkdtemp(prefix="sources-")
    else:
        duplicate_destination = options.target
        if os.path.exists(duplicate_destination):
            raise Exception(
                "Destination directory '{}' exists!".format(duplicate_destination)
            )
    _logger.info("Cloning in %s", duplicate_destination)
    duplicate_repo: Repo
    try:
        if not options.copy_only:
            cvs_kind = options.cvs_kind
            if cvs_kind is None:
                cvs_kind = repo.kind
            duplicate_repo = clone_from(cvs_kind, duplicate_url, duplicate_destination)
            ancestor_branches = repo.ancestor_branches
            duplicate_repo.pre_switch_rm_files(duplicate_destination)
            # update clone to same branch
            duplicate_repo.switch(branch, ancestor_branches)
            duplicate_repo.rm_files(duplicate_destination)
        _copy_files(conf, duplicate_destination)
        if not options.copy_only:
            duplicate_repo.addremove()
            # commit
            if duplicate_repo.has_changes:
                _logger.info("Committing changes")
                duplicate_repo.commit(
                    options.message, repo.current_author_date, repo.current_author
                )
            else:
                _logger.info("No change, no commit")
            # retag
            if options.apply_tags:
                if tags:
                    _logger.info("Applying %d tag(s)", len(tags))
                    duplicate_repo.tag(tags, options.force_tags)
                else:
                    _logger.info("No tag to apply")
            if options.push:
                # push
                _logger.info("Pushing duplicate")
                duplicate_repo.push()
    finally:
        # clean up cloned repository
        if options.clean_up:
            _logger.info("Cleaning up duplicate sources")
            shutil.rmtree(duplicate_destination)


def _copy_files(configuration: Configuration, target: str):
    packages = list(
        package
        for package, options in configuration.python_packages.items()
        if not options["exclude_in_sources"]
    )
    _logger.info(
        "Copying files (%d modules, %d packages)",
        len(configuration.sources_modules),
        len(packages),
    )
    # rsync content inside the other directory
    cmd = [
        "rsync",
        "--cvs-exclude",
        # XXX ignore tous les .hgtags et .hgignore et .hg/*
        #     à refaire sur les sources copiées comme ici ?
        "--exclude=.hg*",
        "--exclude=.git*",
        "--exclude='*.pyc'",
        "--exclude=.drone.yml",
        "--exclude=.gitlab-ci.yml",
        "--links",
        "--copy-unsafe-links",
        "--recursive",
        "--relative",
        "--times",
        *configuration.sources_modules,
        *configuration.dependencies,
        *configuration.other_sources,
        *packages,
        target,
    ]
    _logger.debug(" ".join(cmd))
    call(cmd)
    cvs_parents = set()
    for package in packages + configuration.sources_modules:
        result = _find_cvs_parent(package)
        if result is not None:
            cvs_parent, _package_path_in_target = result
            cvs_parents.add(cvs_parent)
    for element in configuration.dependencies + configuration.other_sources:
        directory = element if os.path.isdir(element) else os.path.dirname(element)
        if directory not in ("", "."):
            result = _find_cvs_parent(package)
            if result is not None:
                cvs_parent, _package_path_in_target = result
                if os.path.exists(os.path.join(target, cvs_parent)):
                    cvs_parents.add(cvs_parent)
    for cvs_parent in cvs_parents:
        get_repo(cvs_parent).write_archival_txt(os.path.join(target, cvs_parent))


if __name__ == "__main__":
    sys.exit(main())
